from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.runtime_context.InterpContext import InterpContext


class NoopInterp(AbstractInterp):
    """
    Interpreter which does nothing

    Actually, it:
    *   stops processing of the command line on itself (as `next_interp` returns `None` by default)
    *   sets `arbitrary_comment` based on config
    """

    def __init__(
        self,
        interp_factory_id,
        config_dict: dict,
        interp_ctx: InterpContext,
    ):
        super().__init__(
            interp_factory_id,
            config_dict,
            interp_ctx,
        )
        self.arbitrary_comment: str = config_dict["arbitrary_comment"]

    def try_iterate(self) -> InterpStep:
        return InterpStep.NextInterp



from marshmallow import Schema, RAISE, fields, post_load

from argrelay.misc_helper import get_config_path
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.schema_config_core_client.ConnectionConfigSchema import connection_config_desc

__comment___ = "__comment__"
connection_config_ = "connection_config"
use_local_requests_ = "use_local_requests"


class ClientConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    # Allow this field in JSON (otherwise schema validation fails):
    __comment__ = fields.String()

    # Serve requests from local data or send to server:
    use_local_requests = fields.Boolean()

    connection_config = fields.Nested(connection_config_desc.dict_schema)

    @post_load
    def make_object(self, input_dict, **kwargs):
        return ClientConfig(
            use_local_requests = input_dict[use_local_requests_],
            connection_config = input_dict[connection_config_],
        )


client_config_desc = TypeDesc(
    dict_schema = ClientConfigSchema(),
    ref_name = ClientConfigSchema.__name__,
    dict_example = {
        use_local_requests_: False,
        connection_config_: connection_config_desc.dict_example,
    },
    default_file_path = get_config_path("argrelay.client.json"),
)



from dataclasses import dataclass, field

import yaml
from marshmallow import Schema, ValidationError


@dataclass(frozen = True)
class TypeDesc:
    """
    Type descriptor: schema, example, etc.
    """

    dict_schema: Schema = field()
    ref_name: str = field()
    dict_example: dict = field()
    default_file_path: str = field()

    def from_default_file(self):
        return self.from_yaml_file(self.default_file_path)

    def from_yaml_file(self, file_path: str):
        return self.from_input_dict(yaml.safe_load(open(file_path)))

    def from_yaml_str(self, yaml_str: str):
        return self.from_input_dict(yaml.safe_load(yaml_str))

    def from_input_dict(self, input_dict: dict):
        return self.dict_schema.load(input_dict)

    def validate_dict(self, input_dict: dict):
        validation_errors = self.dict_schema.validate(input_dict)
        if validation_errors:
            raise ValidationError(validation_errors)



from argrelay.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay.handler_response.AbstractClientResponseHandler import AbstractClientResponseHandler
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_client.AbstractClientCommand import AbstractClientCommand
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.server_spec.CallContext import CallContext


class AbstractLocalClientCommand(AbstractClientCommand):

    def __init__(
        self,
        call_ctx: CallContext,
        server_config: ServerConfig,
        local_server: LocalServer,
        request_handler: AbstractServerRequestHandler,
        response_handler: AbstractClientResponseHandler,
    ):
        super().__init__(
            response_handler,
        )
        self.call_ctx: CallContext = call_ctx
        self.server_config: ServerConfig = server_config
        self.local_server: LocalServer = local_server
        self.request_handler: AbstractServerRequestHandler = request_handler
        self.response_dict: dict
        self.interp_ctx: InterpContext

    def execute_command(
        self,
    ):
        self.response_dict = self.request_handler.handle_request(self.call_ctx)
        ElapsedTime.measure("before_sending_response")
        self.response_handler.handle_response(self.response_dict)
        ElapsedTime.measure("after_handle_response")

        # `AbstractLocalClientCommand`'s feature: expose `InterpContext` used by the server:
        self.interp_ctx = self.request_handler.interp_ctx



from argrelay.client_spec.ShellContext import ShellContext
from argrelay.relay_client import AbstractClientCommand
from argrelay.server_spec.CallContext import CallContext


class AbstractClientCommandFactory:

    def create_command_by_server_path(
        self,
        call_ctx: CallContext,
    ) -> AbstractClientCommand:
        pass

    def create_command(
        self,
        shell_ctx: ShellContext,
    ):
        call_ctx = CallContext.from_shell_context(
            shell_ctx,
        )
        return self.create_command_by_server_path(call_ctx)



from __future__ import annotations

from argrelay.custom_integ.ServiceDelegator import redirect_to_no_func_error
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.enum_desc.TermColor import TermColor
from argrelay.plugin_delegator.AbstractDelegator import get_data_envelopes
from argrelay.plugin_delegator.InterceptDelegator import InterceptDelegator
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.relay_server.QueryEngine import populate_query_dict
from argrelay.runtime_context.InterpContext import function_envelope_ipos_, InterpContext
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.schema_config_interp.DataEnvelopeSchema import envelope_id_
from argrelay.schema_config_interp.SearchControlSchema import search_control_desc

subsequent_function_envelope_ipos_ = function_envelope_ipos_ + 1


# TODO: It inherits `InterceptDelegator`, but it makes more sense to have common base class instead.
class HelpDelegator(InterceptDelegator):

    def run_interp_control(
        self,
        curr_interp: AbstractInterp,
    ) -> str:
        # TODO: This must be special interpreter which is configured only to search functions (without their args).
        return super().run_interp_control(curr_interp)

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_funct_found(), "the (first) function envelope must be found"

        if interp_ctx.curr_container_ipos >= subsequent_function_envelope_ipos_:
            subsequent_function_container = interp_ctx.envelope_containers[(
                subsequent_function_envelope_ipos_
            )]
            query_dict = populate_query_dict(subsequent_function_container)
            delegator_plugin_instance_id = HelpDelegator.__name__

            custom_plugin_data = search_control_desc.dict_schema.dump(subsequent_function_container.search_control)

            invocation_input = InvocationInput(
                all_tokens = interp_ctx.parsed_ctx.all_tokens,
                consumed_tokens = interp_ctx.consumed_tokens,
                delegator_plugin_entry = local_server.server_config.plugin_dict[delegator_plugin_instance_id],
                data_envelopes = (
                    # Envelope of `SpecialFunc.help_func`:
                    get_data_envelopes(interp_ctx)[:subsequent_function_envelope_ipos_]
                    +
                    # These must be function envelopes found via query:
                    local_server.get_query_engine().query_data_envelopes(query_dict)
                ),
                custom_plugin_data = custom_plugin_data,
            )
            return invocation_input
        else:
            return redirect_to_no_func_error(
                interp_ctx,
                local_server.server_config,
            )

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        search_control: SearchControl = search_control_desc.dict_schema.load(invocation_input.custom_plugin_data)
        if invocation_input.data_envelopes[function_envelope_ipos_][envelope_id_] == SpecialFunc.help_func.name:
            for data_envelope in invocation_input.data_envelopes[subsequent_function_envelope_ipos_:]:

                # Print fields from search control:
                for keys_to_types in search_control.keys_to_types_list:
                    # There should be only one (key, value) pair:
                    key_name = next(iter(keys_to_types))
                    type_name = keys_to_types[key_name]
                    print(f"{data_envelope[type_name]}", end = " ")

                # TODO: perform color control only if the output is a terminal:

                print(TermColor.BRIGHT_BLACK.value, end = "")
                print("#", end = " ")
                print(TermColor.RESET.value, end = "")

                if envelope_id_ in data_envelope:
                    print(TermColor.BRIGHT_BLACK.value, end = "")
                    print(f"{data_envelope[envelope_id_]}", end = " ")
                    print(TermColor.RESET.value, end = "")
                else:
                    print(TermColor.DARK_RED.value, end = "")
                    print(f"[no `{envelope_id_}`]", end = " ")
                    print(TermColor.RESET.value, end = "")

                print(TermColor.DARK_GREEN.value, end = "")
                print("#", end = " ")
                print(TermColor.RESET.value, end = "")

                if ReservedArgType.HelpHint.name in data_envelope:
                    print(TermColor.DARK_GREEN.value, end = "")
                    print(
                        f"{data_envelope[ReservedArgType.HelpHint.name]}",
                        end = " ",
                    )
                    print(TermColor.RESET.value, end = "")
                else:
                    print(TermColor.DARK_RED.value, end = "")
                    print(f"[no `{ReservedArgType.HelpHint.name}`]", end = " ")
                    print(TermColor.RESET.value, end = "")

                print()



from argrelay.schema_request.CallContextSchema import call_context_desc
from argrelay.schema_response.ArgValuesSchema import arg_values_desc, arg_values_
from argrelay.server_spec.server_data_schema import get_schema_definitions

spec_data = {
    "summary": "Propose arg values for the given command line and cursor position",
    "consumes": [
        "application/json",
    ],
    "produces": [
        "application/json",
        "text/plain",
    ],
    "parameters": [
        {
            "name": call_context_desc.ref_name,
            "in": "body",
            "required": True,
            "schema": {
                "$ref": "#/definitions/" + call_context_desc.ref_name,
            },
        },
    ],
    "responses": {
        "200": {
            "description": "List of proposed arg values",
            "schema": {
                "$ref": "#/definitions/" + arg_values_desc.ref_name,
            },
            "examples": {
                "application/json": arg_values_desc.dict_example,
                "text/plain": "\n".join(arg_values_desc.dict_example[arg_values_]),
            },
        },
    },
    "definitions": get_schema_definitions([
        call_context_desc.ref_name,
        arg_values_desc.ref_name,
    ]),
}



from argrelay.enum_desc.GlobalArgType import GlobalArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.schema_config_interp.FuncArgsInterpConfigSchema import (
    FuncArgsInterpConfigSchema,
    function_search_control_,
    function_init_control_,
)
from argrelay.schema_config_interp.InitControlSchema import init_types_to_values_
from argrelay.schema_config_interp.SearchControlSchema import envelope_class_, keys_to_types_list_


class DemoInterpFactoryConfigSchema(FuncArgsInterpConfigSchema):
    pass


demo_interp_factory_config_example = {
    function_search_control_: {
        envelope_class_: ReservedEnvelopeClass.ClassFunction.name,
        keys_to_types_list_: [
            {
                "category": GlobalArgType.FunctionCategory.name,
            },
            {
                "action": GlobalArgType.ActionType.name,
            },
            {
                "object": GlobalArgType.ObjectSelector.name,
            },
        ],
    },
    function_init_control_: {
        init_types_to_values_: {
            GlobalArgType.FunctionCategory.name: "internal",
            GlobalArgType.ActionType.name: "intercept",
            GlobalArgType.ObjectSelector.name: "func",
        },
    },
}

demo_interp_factory_config_desc = TypeDesc(
    dict_schema = DemoInterpFactoryConfigSchema(),
    ref_name = DemoInterpFactoryConfigSchema.__name__,
    dict_example = demo_interp_factory_config_example,
    default_file_path = "",
)



from argrelay.handler_response.AbstractClientResponseHandler import AbstractClientResponseHandler


class ProposeArgValuesClientResponseHandler(AbstractClientResponseHandler):

    def __init__(
        self,
    ):
        super().__init__(
        )

    # TODO: this does not look correct: server sends plain text (new-line-separated suggestions):
    def handle_response(self, response_dict: dict):
        print(response_dict["arg_values"])



from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_context.ParsedContext import ParsedContext
from argrelay.server_spec.CallContext import CallContext


class AbstractServerRequestHandler:

    def __init__(
        self,
        local_server: LocalServer,
    ):
        self.local_server: LocalServer = local_server
        self.interp_ctx: InterpContext

    def handle_request(
        self,
        call_ctx: CallContext,
    ) -> dict:
        raise NotImplementedError

    def interpret_command(
        self,
        local_server: LocalServer,
        call_ctx: CallContext,
    ) -> None:
        parsed_ctx = ParsedContext.from_instance(call_ctx)
        # TODO: Split server_config and static_data (both top level, not config including data):
        self.interp_ctx = InterpContext(
            parsed_ctx = parsed_ctx,
            interp_factories = local_server.server_config.interp_factories,
            action_delegators = local_server.server_config.action_delegators,
            query_engine = local_server.get_query_engine(),
            help_hint_cache = local_server.help_hint_cache,
        )
        self.interp_ctx.interpret_command(local_server.server_config.static_data.first_interp_factory_id)
        self.interp_ctx.print_debug()



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper import eprint
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue

indent_size = 2


@dataclass
class EnvelopeContainer:
    """
    Combination of relevant config, user input and `data_envelope` from the store.
    """

    search_control: SearchControl = field(default_factory = lambda: SearchControl())
    """
    A specs how to query `data_envelope` for this `EnvelopeContainer`.
    """

    data_envelope: dict = field(default = None)
    """
    If the last query finds (unique) single result,
    it contains `data_envelope` based on `search_control` and command line input.
    """

    found_count: int = field(default = 0)
    """
    Counter of `data_envelope`-s found in the last query.
    """

    # TODO: Maybe rename to `context_types_to_values`?
    # TODO: Part of (or actually is?) `args_context`: FS_62_25_92_06:
    assigned_types_to_values: dict[str, AssignedValue] = field(default_factory = lambda: {})
    """
    All assigned args (from interpreted tokens) mapped as type:value which belong to `data_envelope`.
    """

    # TODO: Part of (or not?) `args_context` (FS_62_25_92_06) to support FS_13_51_07_97 (single out implicit values):
    remaining_types_to_values: dict[str, list[str]] = field(default_factory = lambda: {})
    """
    All arg values per arg type left as options to match arg value given on the command line.

    When arg value from command line matches one of the values, this arg type moves to `assigned_types_to_values`.
    """

    def populate_implicit_arg_values(self):
        """
        When `data_envelope` is singled out, all remaining `arg_type`-s become `ArgSource.ImplicitValue`.

        # Implements: FS_13_51_07_97
        """
        # Filter as in: FS_31_70_49_15:
        for arg_type in self.search_control.keys_to_types_dict.values():
            if arg_type not in self.assigned_types_to_values:
                if arg_type in self.remaining_types_to_values:
                    arg_values = self.remaining_types_to_values[arg_type]
                    if len(arg_values) == 1:
                        del self.remaining_types_to_values[arg_type]
                        self.assigned_types_to_values[arg_type] = AssignedValue(
                            arg_values[0],
                            ArgSource.ImplicitValue,
                        )

    @staticmethod
    def describe_data(envelope_containers: list[EnvelopeContainer]):
        eprint()
        # TODO: indicate token position - perform opposite of `parse_line_and_cpos`
        is_first_missing_found: bool = False
        for envelope_container in envelope_containers:
            eprint(f"{envelope_container.search_control.envelope_class}: {envelope_container.found_count}")

            for key_to_type_dict in envelope_container.search_control.keys_to_types_list:
                arg_key = next(iter(key_to_type_dict))
                arg_type = key_to_type_dict[arg_key]

                if arg_type in envelope_container.assigned_types_to_values:
                    eprint(" " * indent_size, end = "")
                    if envelope_container.assigned_types_to_values[arg_type].arg_source == ArgSource.ExplicitPosArg:
                        eprint(TermColor.BRIGHT_BLUE.value, end = "")
                    else:
                        eprint(TermColor.DARK_GREEN.value, end = "")
                    eprint(f"{arg_type}:", end = "")
                    eprint(
                        f" {envelope_container.assigned_types_to_values[arg_type].arg_value} " +
                        f"[{envelope_container.assigned_types_to_values[arg_type].arg_source.name}]",
                        end = ""
                    )
                    eprint(TermColor.RESET.value, end = "")
                elif arg_type in envelope_container.remaining_types_to_values:
                    eprint(" " * indent_size, end = "")
                    eprint(TermColor.BRIGHT_YELLOW.value, end = "")
                    if not is_first_missing_found:
                        eprint(f"*{arg_type}:", end = "")
                        is_first_missing_found = True
                    else:
                        eprint(f"{arg_type}:", end = "")
                    eprint(f" ?", end = "")
                    eprint(TermColor.RESET.value, end = "")
                    eprint(
                        f" {' '.join(envelope_container.remaining_types_to_values[arg_type])}",
                        end = ""
                    )
                else:
                    # The arg type is in the remaining but data have no arg values to suggest.
                    # Such arg types are shown because they are part of `search_control`.
                    # But they cannot be specified for current situation, otherwise, if already no data,
                    # any arg value assigned to such arg type would return no results.
                    eprint(" " * indent_size, end = "")
                    eprint(TermColor.BRIGHT_BLACK.value, end = "")
                    eprint(f"{arg_type}:", end = "")
                    eprint(" [none]", end = "")
                    eprint(TermColor.RESET.value, end = "")

                eprint()



from __future__ import annotations

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.misc_helper.AbstractPlugin import AbstractPlugin
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.InterpContext import function_envelope_ipos_
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
    envelope_id_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    search_control_list_,
)
from argrelay.schema_config_interp.SearchControlSchema import search_control_desc


def get_data_envelopes(
    interp_ctx: "InterpContext",
):
    return [envelope_container.data_envelope for envelope_container in interp_ctx.envelope_containers]


def get_func_name_from_container(
    interp_ctx: "InterpContext",
):
    func_data_envelope = interp_ctx.envelope_containers[(
        interp_ctx.curr_interp.base_envelope_ipos + function_envelope_ipos_
    )].data_envelope
    func_name = func_data_envelope[envelope_id_]
    return func_name


def get_func_name_from_envelope(
    data_envelopes: list[dict],
):
    func_data_envelope = data_envelopes[(
        function_envelope_ipos_
    )]
    func_name = func_data_envelope[envelope_id_]
    return func_name


class AbstractDelegator(AbstractPlugin):
    """
    `DelegatorPlugin` implements two sides:
    *   server-side `invoke_control` prepares data in :class:`InvocationInput` (whatever is necessary)
    *   client-side `invoke_action` uses data in :class:`InvocationInput` to execute the action anyway it can

    To simplify reasoning, ensure that both server-side and client-side:
    *   have access to the same code (non-breaking differences are possible, but same code version ensures that)
    *   share data only via :class:`InvocationInput`
    """

    def run_search_control(
        self,
        function_data_envelope: dict,
    ) -> list[SearchControl]:
        return self.extract_search_control_from_function_data_envelope(function_data_envelope)

    def run_init_control(
        self,
        envelope_containers: list[EnvelopeContainer],
        curr_container_ipos: int,
    ):
        self.init_envelope_class(
            envelope_containers,
            curr_container_ipos,
        )

    def run_fill_control(
        self,
        interp_ctx: "InterpContext",
    ):
        pass

    def run_interp_control(
        self,
        curr_interp: "AbstractInterp",
    ):
        pass

    def run_invoke_control(
        self,
        interp_ctx: "InterpContext",
        local_server: "LocalServer",
    ) -> InvocationInput:
        """
        Server-side entry point.
        The plugin instance is used by server after `AbstractPlugin.activate_plugin`.
        """
        pass

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        """
        Client-side (static) entry point.
        The plugin instance is used by client:
        *   without instantiating
        *   without providing `AbstractPlugin.config_dict`
        *   without calling `AbstractPlugin.activate_plugin`
        """
        pass

    @staticmethod
    def extract_search_control_from_function_data_envelope(function_data_envelope: dict):
        return [
            search_control_desc.dict_schema.load(search_control_dict)
            for search_control_dict in function_data_envelope[instance_data_][search_control_list_]
        ]

    @staticmethod
    def init_envelope_class(
        envelope_containers: list[EnvelopeContainer],
        curr_container_ipos: int,
    ):
        curr_container = envelope_containers[curr_container_ipos]
        curr_container.assigned_types_to_values[
            ReservedArgType.EnvelopeClass.name
        ] = AssignedValue(
            curr_container.search_control.envelope_class,
            ArgSource.InitValue,
        )



from dataclasses import dataclass, field


@dataclass
class MongoServerConfig:
    # Example:
    # "test"
    database_name: str = field()

    start_server: bool = field()

    # Example:
    # "~/argrelay.git/temp/mongo/mongodb-linux-x86_64-rhel80-6.0.3/bin/mongod --dbpath ~/argrelay.git/temp/mongo/data"
    server_start_command: str = field()



from enum import Enum


class ServerAction(Enum):
    """
    Defines both:
    *   `ServerAction` in REST API URL path for server request (via enum item value)
    *   `ServerAction` in server request payload (via enum item name)
    """

    DescribeLineArgs = "/describe_line_args/"
    ProposeArgValues = "/propose_arg_values/"
    RelayLineArgs = "/relay_line_args/"



from marshmallow import Schema, fields, RAISE, post_load, pre_dump

from argrelay.enum_desc.PluginType import PluginType
from argrelay.misc_helper import ensure_value_is_enum
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.plugin_loader.NoopLoader import NoopLoader
from argrelay.runtime_data.PluginEntry import PluginEntry

plugin_config_ = "plugin_config"
plugin_module_name_ = "plugin_module_name"
plugin_class_name_ = "plugin_class_name"
plugin_type_ = "plugin_type"

_plugin_entry_example = {
    plugin_module_name_: NoopLoader.__module__,
    plugin_class_name_: NoopLoader.__name__,
    plugin_type_: PluginType.LoaderPlugin.name,
    plugin_config_: {
    },
}


class PluginEntrySchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    plugin_module_name = fields.String(
        required = True,
    )

    plugin_class_name = fields.String(
        required = True,
    )

    plugin_type = fields.Enum(
        PluginType,
        by_value = False,
        required = True,
    )

    plugin_config = fields.Dict(
        required = True,
    )

    @pre_dump
    def make_dict(self, input_object: PluginEntry, **kwargs):
        # TODO: figure out to populate all automatically and reduce duplication - this is error-prone:
        if isinstance(input_object, PluginEntry):
            return {
                plugin_module_name_: input_object.plugin_module_name,
                plugin_class_name_: input_object.plugin_class_name,
                plugin_type_: ensure_value_is_enum(input_object.plugin_type, PluginType),
                plugin_config_: input_object.plugin_config,
            }
        else:
            # Assuming it is as dict:
            return {
                plugin_module_name_: input_object[plugin_module_name_],
                plugin_class_name_: input_object[plugin_class_name_],
                plugin_type_: ensure_value_is_enum(input_object[plugin_type_], PluginType),
                plugin_config_: input_object[plugin_config_],
            }

    @post_load
    def make_object(self, input_dict, **kwargs):
        return PluginEntry(
            plugin_module_name = input_dict[plugin_module_name_],
            plugin_class_name = input_dict[plugin_class_name_],
            plugin_type = ensure_value_is_enum(input_dict[plugin_type_], PluginType),
            plugin_config = input_dict[plugin_config_],
        )


plugin_entry_desc = TypeDesc(
    dict_schema = PluginEntrySchema(),
    ref_name = PluginEntrySchema.__name__,
    dict_example = _plugin_entry_example,
    default_file_path = "",
)



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.mongo_data.MongoConfig import MongoConfig
from argrelay.relay_server.GuiBannerConfig import GuiBannerConfig
from argrelay.relay_server.QueryCacheConfig import QueryCacheConfig
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.runtime_data.StaticData import StaticData


@dataclass
class ServerConfig:
    connection_config: ConnectionConfig = field()
    mongo_config: MongoConfig = field()
    query_cache_config: QueryCacheConfig = field()
    gui_banner_config: GuiBannerConfig = field()

    plugin_instance_id_load_list: list[str] = field()
    """
    List of `plugin_instance_id`s in order of loading. Each `plugin_instance_id` is a key into `plugin_dict`.
    """

    plugin_dict: dict[str, PluginEntry] = field()
    """
    Key = `plugin_instance_id`
    """

    static_data: StaticData = field()

    # TODO: Keep this runtime objects in separate (`ServerRuntime`?) class. Ensure/implement ServerConfig dumping on request (for troubleshooting).
    data_loaders: dict[str, "AbstractLoader"] = field(default_factory = lambda: {})
    """
    Entries in `data_loaders` are not directly loaded from config.
    These are plugin instances created during plugin activation.
    """

    # TODO: Keep this runtime objects in separate (`ServerRuntime`?) class. Ensure/implement ServerConfig dumping on request (for troubleshooting).
    interp_factories: dict[str, "AbstractInterpFactory"] = field(default_factory = lambda: {})
    """
    Entries in `interp_factories` are not directly loaded from config.
    These are plugin instances created during plugin activation.
    """

    # TODO: Keep this runtime objects in separate (`ServerRuntime`?) class. Ensure/implement ServerConfig dumping on request (for troubleshooting).
    action_delegators: dict[str, "AbstractDelegator"] = field(default_factory = lambda: {})
    """
    Entries in `action_delegators` are not directly loaded from config.
    These are plugin instances created during plugin activation.
    """



from unittest import TestCase

from argrelay.client_command_local.AbstractLocalClientCommand import AbstractLocalClientCommand
from argrelay.enum_desc.RunMode import RunMode
from argrelay.relay_client import __main__
from argrelay.schema_response.ArgValuesSchema import arg_values_
from argrelay.test_helper import parse_line_and_cpos
from argrelay.test_helper.EnvMockBuilder import EnvMockBuilder


class InOutTestCase(TestCase):

    # TODO: wrap input into Dataclass which in turn can be created via builder (to have ability pre-build defaults for a set of tests).
    def verify_output(
        self,
        test_data,
        test_line,
        comp_type,
        expected_suggestions,
        envelope_ipos_to_expected_assignments,
        delegator_class,
    ):
        (command_line, cursor_cpos) = parse_line_and_cpos(test_line)

        if delegator_class:
            init_env_mock_builder = EnvMockBuilder().set_capture_delegator_invocation_input(delegator_class)
        else:
            init_env_mock_builder = EnvMockBuilder()

        env_mock_builder = (
            init_env_mock_builder
            .set_command_line(command_line)
            .set_cursor_cpos(cursor_cpos)
            .set_comp_type(comp_type)
            .set_test_data_ids_to_load([
                test_data,
            ])
        )
        with env_mock_builder.build():
            command_obj = __main__.main()
            assert isinstance(command_obj, AbstractLocalClientCommand)

            # TODO: Be able to choose to verify output payload instead of server-side internal data
            #       (only available in tests when `LocalClient` and `LocalServer` are used).
            interp_ctx = command_obj.interp_ctx

            # TODO: Currently, this verifier ensures what things exists.
            #       Add a way to ensure what things do not exists.
            #       For example,
            #       * no `data_envelope` number N.
            #       * `data_envelope` does not have field A.
            #       * `data_envelope` does not have value 'x' for field A.

            if envelope_ipos_to_expected_assignments:
                for envelope_ipos in envelope_ipos_to_expected_assignments.keys():
                    # Verify `envelope_container` of `envelope_ipos` exists in the list:
                    self.assertTrue(0 <= envelope_ipos < len(interp_ctx.envelope_containers))

            if RunMode.from_comp_type(comp_type) == RunMode.CompletionMode:
                actual_suggestions = command_obj.response_dict[arg_values_]
                self.assertEqual(expected_suggestions, actual_suggestions)
                self.assertFalse(
                    envelope_ipos_to_expected_assignments,
                    "no assignments expected in `RunMode.CompletionMode`",
                )
                self.assertFalse(
                    delegator_class,
                    "no invocation expected in `RunMode.CompletionMode`",
                )
            elif RunMode.from_comp_type(comp_type) == RunMode.InvocationMode:
                self.assertFalse(
                    expected_suggestions,
                    "no suggestions expected in `RunMode.InvocationMode`",
                )
                for envelope_ipos, expected_assignments in envelope_ipos_to_expected_assignments.items():
                    for arg_type, arg_value in expected_assignments.items():
                        if arg_value is None:
                            self.assertTrue(
                                arg_type not in
                                interp_ctx.envelope_containers
                                [envelope_ipos].assigned_types_to_values
                            )
                        else:
                            self.assertEqual(
                                arg_value,
                                interp_ctx.envelope_containers
                                [envelope_ipos].assigned_types_to_values
                                [arg_type]
                            )



from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper.TypeDesc import TypeDesc

error_message_ = "error_message"

error_code_ = "error_code"


class ErrorDelegatorCustomDataSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    error_message = fields.String(
        required = False,
    )

    error_code = fields.Integer(
        required = False,
    )


error_delegator_stub_custom_data_example = {
    error_message_: "INFO: command executed successfully: demo implementation is a stub",
    error_code_: 0,
}
error_delegator_custom_data_desc = TypeDesc(
    dict_schema = ErrorDelegatorCustomDataSchema(),
    ref_name = ErrorDelegatorCustomDataSchema.__name__,
    dict_example = error_delegator_stub_custom_data_example,
    default_file_path = "",
)



from enum import IntEnum


class ArgSource(IntEnum):
    """
    Specifies how each :class:`AssignedValue` was obtained.

    They are listed in the order of priority:
    *  The first one is the lowest priority = can be overriden by any other `ArgSource.*`
    *  The last one is the highest priority = cannot be overriden by any other `ArgSource.*`
    """

    # Default value selected via `fill_control` (FS_72_40_53_00).
    DefaultValue = 3

    # TODO: Provide detailed description in FS_13_51_07_97?
    # Provided in the envelope singled-out by the search:
    ImplicitValue = 4

    # If command line arg is explicitly provided as positional argument (`TokenType.PosArg`):
    ExplicitPosArg = 5

    # If command line arg is explicitly provided as named argument (`TokenType.KeyArg` + `TokenType.ValArg`):
    ExplicitNamedArg = 6

    # This is pre-assigned value which cannot be changed subsequently (the highest priority).
    # It is a value computed by plugin domain-specific logic with input from:
    # * Env vars.
    # * Config.
    # * Currently found `data_envelope`-s
    InitValue = 7

    def __str__(self):
        return self.name



from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.relay_client.AbstractClientCommandFactory import AbstractClientCommandFactory
from argrelay.runtime_data.ClientConfig import ClientConfig
from argrelay.server_spec.CallContext import CallContext


class RemoteClientCommandFactory(AbstractClientCommandFactory):

    def __init__(
        self,
        client_config: ClientConfig,
    ):
        self.client_config: ClientConfig = client_config

    # noinspection PyMethodMayBeStatic
    def create_command_by_server_path(
        self,
        call_ctx: CallContext,
    ) -> "AbstractRemoteClientCommand":
        if call_ctx.server_action == ServerAction.DescribeLineArgs:
            from argrelay.client_command_remote.DescribeLineArgsRemoteClientCommand import (
                DescribeLineArgsRemoteClientCommand,
            )
            return DescribeLineArgsRemoteClientCommand(
                call_ctx,
                self.client_config.connection_config,
            )
        if call_ctx.server_action == ServerAction.ProposeArgValues:
            from argrelay.client_command_remote.ProposeArgValuesRemoteClientCommand import (
                ProposeArgValuesRemoteClientCommand,
            )
            return ProposeArgValuesRemoteClientCommand(
                call_ctx,
                self.client_config.connection_config,
            )
        if call_ctx.server_action == ServerAction.RelayLineArgs:
            from argrelay.client_command_remote.RelayLineArgsRemoteClientCommand import (
                RelayLineArgsRemoteClientCommand,
            )
            return RelayLineArgsRemoteClientCommand(
                call_ctx,
                self.client_config.connection_config,
            )
        raise RuntimeError



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.client_spec.ShellContext import ShellContext
from argrelay.enum_desc.CompScope import CompScope
from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper import eprint


@dataclass(frozen = True)
class CallContext:
    """
    Immutable input data from client to server
    """

    server_action: ServerAction = field()
    command_line: str = field()
    cursor_cpos: int = field()
    comp_scope: CompScope = field()
    is_debug_enabled: bool = field()

    # TODO: used on server side: to be changed into CallContext server-client-API class
    # TODO: call context should not be created from shell context on server side, it should simply be transferred from client
    @classmethod
    def from_shell_context(
        cls,
        shell_ctx: ShellContext,
    ) -> CallContext:
        server_action: ServerAction = cls.select_server_action(shell_ctx)
        return cls(
            server_action = server_action,
            command_line = shell_ctx.command_line,
            cursor_cpos = shell_ctx.cursor_cpos,
            comp_scope = CompScope.from_comp_type(shell_ctx.comp_type),
            is_debug_enabled = shell_ctx.is_debug_enabled,
        )

    @staticmethod
    def select_server_action(
        shell_ctx: ShellContext,
    ) -> ServerAction:
        if shell_ctx.comp_type == CompType.DescribeArgs:
            return ServerAction.DescribeLineArgs
        if shell_ctx.comp_type == CompType.InvokeAction:
            return ServerAction.RelayLineArgs

        return ServerAction.ProposeArgValues

    def print_debug(
        self,
        end_str: str = "\n",
    ) -> None:
        if not self.is_debug_enabled:
            return
        eprint(TermColor.DEBUG.value, end = "")
        eprint(f"server_action: {self.server_action}", end = " ")
        eprint(f"\"{self.command_line}\"", end = " ")
        eprint(f"cursor_cpos: {self.cursor_cpos}", end = " ")
        eprint(f"comp_scope: {self.comp_scope}", end = " ")
        eprint(TermColor.RESET.value, end = end_str)



"""
Mock environment (env vars, command line args, file access, database, etc.) for `argrelay` client or server
"""

from __future__ import annotations

import contextlib
import dataclasses
import io
import json
import os
import re
import sys
from contextlib import ExitStack
from dataclasses import dataclass, field
from io import StringIO
from unittest import mock

import mongomock
import pkg_resources
import yaml

import argrelay
from argrelay.client_spec.ShellContext import (
    UNKNOWN_COMP_KEY,
    ShellContext,
    COMP_LINE_env_var,
    COMP_POINT_env_var,
    COMP_TYPE_env_var,
    COMP_KEY_env_var,
)
from argrelay.custom_integ.GitRepoLoader import GitRepoLoader
from argrelay.custom_integ.GitRepoLoaderConfigSchema import is_plugin_enabled_
from argrelay.custom_integ.ServiceLoader import ServiceLoader
from argrelay.custom_integ.ServiceLoaderConfigSchema import test_data_ids_to_load_
from argrelay.enum_desc.CallConv import CallConv
from argrelay.enum_desc.CompType import CompType
from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.mongo_data import MongoClientWrapper
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.runtime_context.ParsedContext import ParsedContext
from argrelay.schema_config_core_client.ClientConfigSchema import use_local_requests_, client_config_desc
from argrelay.schema_config_core_server.MongoConfigSchema import mongo_server_, use_mongomock_only_
from argrelay.schema_config_core_server.MongoServerConfigSchema import start_server_
from argrelay.schema_config_core_server.QueryCacheConfigSchema import enable_query_cache_
from argrelay.schema_config_core_server.ServerConfigSchema import (
    mongo_config_,
    server_config_desc,
    plugin_dict_,
    query_cache_config_,
)
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_config_
from argrelay.server_spec.CallContext import CallContext
from argrelay.test_helper.OpenFileMock import OpenFileMock


@dataclass
class EnvMockBuilder:
    """
    All-in-one mock support which sets up the mocks and cleans them up as Python's Context Manager.

    For example:

    *   Mock env vars or `sys.argv` used by Bash to communicate input to argrelay client - see usage of:

        *   `_mock_client_input_in_completion_mode`
        *   `_mock_client_input_in_invocation_mode_with_args`
        *   `_mock_client_input_in_invocation_mode_with_line`

    *   Mock server and client config files - see usage of:

        *   `set_server_config_dict`
        *   `set_client_config_dict`

    *   Capture `stdout` and `stderr` - see usage of:

        *   `set_capture_stdout`
        *   `set_capture_stderr`

    *   Whether client uses `LocalClient`/`LocalServer` or `RemoteClient` with `CustomFlaskApp` - see usage of:

        *   `set_client_config_with_local_server`

    *   Mock MongoDB client - see usage of: `mock_mongo_client`

    *   Simple selection of test data - see usage of: `set_test_data_ids_to_load`

    *   Verifying plugin `InvocationInput` - see usage of: `delegator_plugin_invoke_action_func_path`

    *   ...

    """

    command_line: str = field(default = "")
    _command_line_is_set: bool = field(default = False)

    command_args: list[str] = field(default_factory = lambda: [])
    _command_args_are_set: bool = field(default = False)

    cursor_cpos: int = field(default = -1)
    _cursor_cpos_is_set: bool = field(default = False)

    comp_type: CompType = field(default = CompType.PrefixShown)
    comp_key: str = field(default = UNKNOWN_COMP_KEY)

    _mock_client_input: bool = field(default = True)

    file_mock: OpenFileMock = field(default = OpenFileMock({}))

    client_config_dict: dict = field(default_factory = lambda: load_custom_integ_client_config_dict())
    mock_client_config_file_read: bool = field(default = True)
    is_client_config_with_local_server: bool = field(default = True)

    server_config_dict: dict = field(default_factory = lambda: load_custom_integ_server_config_dict())
    mock_server_config_file_read: bool = field(default = True)
    is_server_config_with_mongo_start: bool = field(default = False)
    enable_demo_git_loader: bool = field(default = False)

    actual_stdout: StringIO = field(default = None)
    capture_stdout: bool = field(default = False)

    actual_stderr: StringIO = field(default = None)
    capture_stderr: bool = field(default = False)

    mock_mongo_client: bool = field(default = True)

    assert_on_close: bool = field(default = True)

    test_data_ids_to_load: list[str] = field(default_factory = lambda: [
        "TD_70_69_38_46",  # no data
    ])

    delegator_plugin_invoke_action_func_path: str = field(default = None)
    invocation_input: InvocationInput = field(default = None)

    enable_query_cache: bool = field(default = True)

    def set_command_line(self, command_line: str):
        """
        Used as input for `RunMode.CompletionMode` because `COMP_LINE` is env var what Bash sets
        """
        self.command_line = command_line
        self._command_line_is_set = True
        return self

    def set_command_args(self, command_args: list[str]):
        """
        Used as input for `RunMode.InvocationMode` because list[str] is what `sys.argv` provides
        """
        self.command_args = command_args
        self._command_args_are_set = True
        return self

    def set_cursor_cpos(self, cursor_cpos: int):
        self.cursor_cpos = cursor_cpos
        self._cursor_cpos_is_set = True
        return self

    def set_comp_type(self, comp_type: CompType):
        self.comp_type = comp_type
        return self

    def set_comp_key(self, comp_key: str):
        self.comp_key = comp_key
        return self

    def set_mock_client_input(self, given_val: bool):
        self._mock_client_input = given_val
        return self

    def set_client_config_dict(self, client_config: dict):
        self.client_config_dict = client_config
        return self

    def get_client_config_json(self):
        return json.dump(self.client_config_dict)

    def set_mock_client_config_file_read(self, given_val: bool):
        self.mock_client_config_file_read = given_val
        return self

    def set_client_config_with_local_server(self, given_val: bool):
        self.is_client_config_with_local_server = given_val
        return self

    def set_server_config_dict(self, server_config: dict):
        self.server_config_dict = server_config
        return self

    def get_server_config_yaml(self):
        return yaml.dump(self.server_config_dict)

    def set_mock_server_config_file_read(self, given_val: bool):
        self.mock_server_config_file_read = given_val
        return self

    def set_server_config_with_mongo_start(self, given_val: bool):
        self.is_server_config_with_mongo_start = given_val
        return self

    def set_enable_demo_git_loader(self, given_val: bool):
        self.enable_demo_git_loader = given_val
        return self

    def set_capture_stdout(self, given_val: bool):
        self.capture_stdout = given_val
        return self

    def set_capture_stderr(self, given_val: bool):
        self.capture_stderr = given_val
        return self

    def set_mock_mongo_client(self, mock_mongo_client: bool):
        self.mock_mongo_client = mock_mongo_client
        return self

    def set_test_data_ids_to_load(self, test_data_ids_to_load: list[str]):
        self.test_data_ids_to_load = test_data_ids_to_load
        return self

    def set_capture_delegator_invocation_input(self, abstract_delegator: AbstractDelegator):
        """
        This func causes `AbstractDelegator.invoke_action` to be mocked to capture `InvocationInput`
        inside `ErrorDelegator.invocation_input` which test can then assert its data.
        """

        self.delegator_plugin_invoke_action_func_path = (
            f"{abstract_delegator.__module__}"
            "."
            f"{abstract_delegator.__name__}"
            "."
            "invoke_action"
        )
        return self

    def set_enable_query_cache(self, given_val: bool):
        self.enable_query_cache = given_val
        return self

    @contextlib.contextmanager
    def mock_file_open(self):
        with mock.patch("builtins.open", self.file_mock.open) as file_mock:
            yield file_mock

    @contextlib.contextmanager
    def build(self):
        """
        Mock resources via "Combining Multiple Context Managers":
        https://rednafi.github.io/digressions/python/2020/03/26/python-contextmanager.html#combining-multiple-context-managers

        Warning:
            The approach below works in tests for setting up and tearing down mocks (judging by behavior),
            but there is no guarantee it is robust under all scenarios.
            For real resource management (when various exceptions raised), it has to be thoroughly tested.
        """

        # Ensure there are no false expectations if conflicting setup is done:
        assert not (self._command_line_is_set and self._command_args_are_set), "both cannot be true"

        if self._command_line_is_set:
            assert self._cursor_cpos_is_set, "setting command line (in CompletionMode) requires cursor cpos"

        if self._command_args_are_set:
            assert not self._cursor_cpos_is_set, "if args are set (in InvocationMode), cursor pos is not set"

        if self.is_client_config_with_local_server:
            # So far, local server is only used for testing (which implies using mocked client config file access).
            # If fails here, for consistency, either enable client config file mocking or disable local server.
            assert self.mock_client_config_file_read

        if self.is_server_config_with_mongo_start:
            assert self.mock_server_config_file_read

        if self.enable_demo_git_loader:
            assert self.mock_client_config_file_read

        if self.enable_query_cache != self.server_config_dict[query_cache_config_][enable_query_cache_]:
            assert self.mock_server_config_file_read

        if self.mock_client_config_file_read:
            self.client_config_dict[use_local_requests_] = self.is_client_config_with_local_server
            self.file_mock.path_to_data[client_config_desc.default_file_path] = json.dumps(self.client_config_dict)

        if self.mock_server_config_file_read:
            """
            Change server config data, then mock file access to return that data for tests.
            """

            self.server_config_dict[mongo_config_][mongo_server_][
                start_server_
            ] = self.is_server_config_with_mongo_start
            plugin_entry = self.server_config_dict[plugin_dict_][GitRepoLoader.__name__]
            plugin_entry[plugin_config_][is_plugin_enabled_] = self.enable_demo_git_loader

            plugin_entry = self.server_config_dict[plugin_dict_][ServiceLoader.__name__]
            plugin_entry[plugin_config_][test_data_ids_to_load_] = self.test_data_ids_to_load

            self.server_config_dict[query_cache_config_][enable_query_cache_] = self.enable_query_cache

            self.file_mock.path_to_data[server_config_desc.default_file_path] = yaml.dump(self.server_config_dict)

        with ExitStack() as exit_stack:

            # noinspection PyListCreation
            yield_list = []

            # Always mock file access - whether file data or mocked data is given depends on the config:
            yield_list.append(exit_stack.enter_context(self.mock_file_open()))

            if (
                self.mock_mongo_client
                and
                # If `mongomock` is already used, no need to mock MongoDB:
                not self.server_config_dict[mongo_config_][use_mongomock_only_]
            ):
                yield_list.append(exit_stack.enter_context(_mongo_client_mock_manager()))

            if self._mock_client_input:
                if CallConv.from_comp_type(self.comp_type) == CallConv.EnvVarsConv:
                    # TODO: make explicit function "mock_client_input_in_env_vars" with all three args required.
                    yield_list.append(exit_stack.enter_context(
                        _mock_client_input_in_env_vars(
                            self.command_line,
                            self.cursor_cpos,
                            self.comp_type,
                        )
                    ))
                elif CallConv.from_comp_type(self.comp_type) == CallConv.CliArgsConv:
                    # TODO: make explicit function "mock_client_input_in_cli_args" with just command_args.
                    if self._command_args_are_set:
                        # TODO: do not branch here, branch on mock setup (in client tests) to make it explicit/conscious that InvocationMode is not about command_line, but command_args.
                        yield_list.append(exit_stack.enter_context(
                            _mock_client_input_in_invocation_mode_with_args(
                                self.command_args,
                            )
                        ))
                    elif self._command_line_is_set:
                        # TODO: do not branch here, branch on mock setup (in client tests) to make it explicit/conscious that InvocationMode is not about command_line, but command_args.
                        yield_list.append(exit_stack.enter_context(
                            _mock_client_input_in_invocation_mode_with_line(
                                self.command_line,
                            )
                        ))
                    else:
                        raise RuntimeError
                else:
                    raise RuntimeError

            if self.capture_stdout:
                self.actual_stdout = io.StringIO()
                yield_list.append(exit_stack.enter_context(_mock_stdout(self.actual_stdout)))

            if self.capture_stderr:
                self.actual_stderr = io.StringIO()
                yield_list.append(exit_stack.enter_context(_mock_stderr(self.actual_stderr)))

            if self.assert_on_close:
                yield_list.append(exit_stack.enter_context(self.assert_all_cm()))

            if self.delegator_plugin_invoke_action_func_path:
                yield_list.append(exit_stack.enter_context(
                    _mock_delegator_plugin(self.delegator_plugin_invoke_action_func_path)
                ))

            yield yield_list

    @contextlib.contextmanager
    def assert_all_cm(self):
        try:
            yield
        finally:
            if self.mock_client_config_file_read:
                self.assert_client_config_read()
            if self.mock_server_config_file_read:
                self.assert_server_config_read()

    def assert_client_config_read(self):
        self.assert_file_read(client_config_desc.default_file_path)

    def assert_server_config_read(self):
        self.assert_file_read(server_config_desc.default_file_path)

    def assert_file_read(self, file_path: str):
        """
        Ensures that mocked file was actually accessed.

        If fails here, it means either/or:
        *   test setup was over-mocked (e.g. see `mock_server_config_file_read`, `mock_client_config_file_read`)
        *   test did not hit functionality that is supposed to access the file
        """
        self.file_mock.path_to_mock[file_path].assert_called_with(file_path)


@dataclass
class ServerOnlyEnvMockBuilder(EnvMockBuilder):
    """
    Used in tests where client code is not invoked.
    """

    def __init__(
        self,
    ):
        super().__init__()
        # TODO: enable validation that client code is not invoked:
        # For server-only test, client config file read should not happen.
        # Disable expectations:
        self.set_mock_client_config_file_read(False)
        # For server-only test, client code is not used, but if it is, try to fail via REST API:
        self.set_client_config_with_local_server(False)
        # For server-only test, client input mocking is not required:
        self.set_mock_client_input(False)


@dataclass
class LiveServerEnvMockBuilder(EnvMockBuilder):
    """
    Used in tests where client talks to some live server.

    The purpose of this mock is to test server data (rather than code).
    """

    def __init__(
        self,
    ):
        super().__init__()
        # TODO: enable validation that client code is not invoked:
        # For live server test, server config file read should not happen by in-process server code.
        # Disable expectations:
        self.set_mock_server_config_file_read(False)
        # For live server test, running local server contradicts with the purpose:
        self.set_client_config_with_local_server(False)


@contextlib.contextmanager
def _mongo_client_mock_manager():
    get_mongo_client_orig = MongoClientWrapper.get_mongo_client
    target_name = get_mongo_client_orig.__module__ + "." + get_mongo_client_orig.__name__
    with mock.patch(target_name) as get_mongo_client_mock:
        get_mongo_client_mock.return_value = mongomock.MongoClient()
        yield get_mongo_client_mock


@contextlib.contextmanager
def _mock_client_input_in_env_vars(command_line: str, cursor_cpos: int, comp_type: CompType):
    with mock.patch.dict(os.environ, {
        COMP_LINE_env_var: command_line,
        COMP_POINT_env_var: str(cursor_cpos),
        COMP_TYPE_env_var: str(comp_type.value),
        COMP_KEY_env_var: UNKNOWN_COMP_KEY,
    }) as env_mock:
        yield env_mock


def _mock_client_input_in_invocation_mode_with_line(command_line: str):
    command_args = re.compile(SpecialChar.TokenDelimiter.value).split(command_line)
    return _mock_client_input_in_invocation_mode_with_args(command_args)


@contextlib.contextmanager
def _mock_client_input_in_invocation_mode_with_args(command_args: list[str]):
    with mock.patch.object(sys, "argv", command_args) as argv_mock:
        yield argv_mock


@contextlib.contextmanager
def _mock_stdout(stdout_f):
    with contextlib.redirect_stdout(stdout_f) as stdout_mock:
        yield stdout_mock


@contextlib.contextmanager
def _mock_stderr(stderr_f):
    with contextlib.redirect_stderr(stderr_f) as stderr_mock:
        yield stderr_mock


@contextlib.contextmanager
def _mock_delegator_plugin(path_to_invoke_action):
    with mock.patch(path_to_invoke_action, capture_invocation_input) as mock_static:
        yield mock_static


def capture_invocation_input(invocation_input: InvocationInput):
    """
    This body substitutes (mocks) `invoke_action` func in `DelegatorPlugin`-s.

    Instead of executing func logic, it only captures its input for verifications in tests.
    """
    EnvMockBuilder.invocation_input = dataclasses.replace(invocation_input)


def load_custom_integ_server_config_dict() -> dict:
    test_server_config_path = _get_resource_path("sample_conf/argrelay.server.yaml")
    with open(test_server_config_path) as f:
        server_config_dict = yaml.safe_load(f)
    return server_config_dict


def load_custom_integ_client_config_dict() -> dict:
    test_client_config_path = _get_resource_path("sample_conf/argrelay.client.json")
    with open(test_client_config_path) as f:
        client_config_dict = json.load(f)
    return client_config_dict


def _get_resource_path(rel_path: str):
    # Composing path to resource this way keeps its base directory always at this relative path:
    test_server_config_path = pkg_resources.resource_filename(argrelay.__name__, rel_path)
    return test_server_config_path


def default_test_parsed_context(
    command_line: str,
    cursor_cpos: int,
) -> ParsedContext:
    return ParsedContext.from_instance(
        default_test_input_context(
            command_line,
            cursor_cpos,
        ),
    )


def default_test_input_context(
    command_line: str,
    cursor_cpos: int,
) -> CallContext:
    return CallContext.from_shell_context(ShellContext(
        command_line = command_line,
        cursor_cpos = cursor_cpos,
        comp_type = CompType.PrefixShown,
        comp_key = UNKNOWN_COMP_KEY,
        is_debug_enabled = False,
    ))



from dataclasses import dataclass, field

from argrelay.enum_desc.PluginType import PluginType


@dataclass(frozen = True)
class PluginEntry:
    plugin_module_name: str = field()
    plugin_class_name: str = field()
    plugin_type: PluginType = field()
    plugin_config: dict = field()



from __future__ import annotations

from enum import Enum, auto

from argrelay.enum_desc.CompType import CompType


class RunMode(Enum):
    """
    Note: `RunMode` became completely implied and not necessary information for logic (derived from other values).

    See also `CallConv`.

    `RunMode` originates on the client side because of different user actions:
    *   `CompletionMode`: user hits `Tab` trying to complete command line arg.
    *   `InvocationMode`: user hits `Enter` trying to execute the command line.
    Request to `argrelay` server indicates the two different modes:
    *   `CompletionMode` - when shell asks `argrelay` server for suggestions to complete one command line arg.
    *   `InvocationMode` - when already executed command uses `argrelay` to determine values for all args.
    In other words:
    *   `CompletionMode` = (proposing options) collects all possible value options for the single arg under cursor.
    *   `InvocationMode` = (exercising options) selects only one value for specific arg, but does it for every arg.
    Also:
    *   `CompletionMode` excludes tangent token ("touched" by the cursor) as input to provide completion for it.
    *   `InvocationMode` includes tangent token as input arg to execute selected function.
    """

    CompletionMode = auto()
    InvocationMode = auto()

    def __str__(self):
        return self.name

    @classmethod
    def from_comp_type(
        cls,
        comp_type: CompType,
    ) -> RunMode:
        """
        Derive `RunMode` on client side from `CompType`.
        """

        if comp_type == CompType.InvokeAction:
            return RunMode.InvocationMode
        else:
            return RunMode.CompletionMode



from marshmallow import Schema, RAISE, fields, post_load, validates_schema, ValidationError

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.schema_config_interp.DataEnvelopeSchema import data_envelope_desc, mongo_id_
from argrelay.schema_config_interp.SearchControlSchema import (
    search_control_desc,
)
from argrelay.schema_response.AssignedValueSchema import assigned_value_desc
from argrelay.schema_response.FilteredDict import FilteredDict

"""
Schema for :class:`EnvelopeContainer`
"""

search_control_ = "search_control"
data_envelope_ = "data_envelope"
found_count_ = "found_count"
assigned_types_to_values_ = "assigned_types_to_values"
remaining_types_to_values_ = "remaining_types_to_values"


class EnvelopeContainerSchema(Schema):
    class Meta:
        unknown = RAISE
        ordered = True

    search_control = fields.Nested(
        search_control_desc.dict_schema,
        required = True,
    )

    data_envelope = FilteredDict(
        filtered_keys = [mongo_id_],
        required = True,
        allow_none = True,
    )

    found_count = fields.Integer(
        required = True,
    )

    assigned_types_to_values = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(assigned_value_desc.dict_schema),
        required = True,
    )

    remaining_types_to_values = fields.Dict(
        keys = fields.String(),
        values = fields.List(
            fields.String(),
        ),
        required = True,
    )

    @post_load
    def make_object(self, input_dict, **kwargs):
        return EnvelopeContainer(
            search_control = input_dict[search_control_],
            data_envelope = input_dict[data_envelope_],
            found_count = input_dict[found_count_],
            assigned_types_to_values = input_dict[assigned_types_to_values_],
            remaining_types_to_values = input_dict[remaining_types_to_values_],
        )

    @validates_schema
    def validate_known(self, input_dict, **kwargs):
        if input_dict[found_count_] > 0:
            if not input_dict[data_envelope_]:
                raise ValidationError(
                    f"`{found_count_}` is `{input_dict[found_count_]}`"
                    f"but `{data_envelope_}` is `{None}`"
                )


envelope_container_desc = TypeDesc(
    dict_schema = EnvelopeContainerSchema(),
    ref_name = EnvelopeContainerSchema.__name__,
    dict_example = {
        search_control_: search_control_desc.dict_example,
        data_envelope_: data_envelope_desc.dict_example,
        found_count_: 1,
        assigned_types_to_values_: {
            "SomeTypeA": assigned_value_desc.dict_example,
        },
        remaining_types_to_values_: {
            "SomeTypeB": [
                "B_value_1",
                "B_value_7",
            ],
            "SomeTypeC": [
                "C_value_3",
                "C_value_9",
            ],
        }
    },
    default_file_path = "",
)



from __future__ import annotations

from argrelay.handler_response.AbstractClientResponseHandler import AbstractClientResponseHandler
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.schema_response.InterpResult import InterpResult
from argrelay.schema_response.InterpResultSchema import interp_result_desc


class DescribeLineArgsClientResponseHandler(AbstractClientResponseHandler):

    def __init__(
        self,
    ):
        super().__init__(
        )

    def handle_response(self, response_dict: dict):
        interp_result: InterpResult = interp_result_desc.dict_schema.load(response_dict)
        ElapsedTime.measure("after_object_creation")
        interp_result.describe_data()



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass(frozen = True)
class StaticData:
    """
    All data which normally does not change between requests
    """

    first_interp_factory_id: str = field(init = True, default = "")

    known_arg_types: list = field(init = True, default_factory = lambda: [])
    """
    TODO: rename to make it explicit that it is indexed fields.
    Field `known_arg_types` lists fields of `data_envelop` - they are used to create MongoDB index.
    """

    data_envelopes: list = field(init = True, default_factory = lambda: [])



from marshmallow import Schema, RAISE, fields, validates_schema, ValidationError

from argrelay.custom_integ.DemoInterpFactory import DemoInterpFactory
from argrelay.misc_helper.TypeDesc import TypeDesc

interp_selector_tree_ = "interp_selector_tree"

default_leaf_ = ""


def validate_tree_node(interp_selector_sub_tree: dict):
    if isinstance(interp_selector_sub_tree, str):
        return
    elif isinstance(interp_selector_sub_tree, dict):
        for sub_tree_node in interp_selector_sub_tree.values():
            validate_tree_node(sub_tree_node)
    else:
        raise ValidationError(f"neither a str nor a dict: {interp_selector_sub_tree}")


class TreePathInterpFactoryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    interp_selector_tree = fields.Dict(
        keys = fields.String(),
        # This is a tree (Dict) of arbitrary depth with leaves being of type String.
        # Ideally, this should be defined either String or nested Dict,
        # but it is unknown how to do it in marshmallow.
        values = fields.Raw(),
        required = True,
    )

    @validates_schema
    def validate_known(self, input_dict, **kwargs):
        interp_selector_tree = input_dict[interp_selector_tree_]
        if not isinstance(interp_selector_tree, dict):
            raise ValidationError(f"not a dict: {interp_selector_tree}")
        validate_tree_node(interp_selector_tree)


tree_path_interp_factory_config_example = {
    interp_selector_tree_: {
        # TODO: This may not be currently configured plugin id:
        "intercept": f"{DemoInterpFactory.__name__}.intercept_func",
        default_leaf_: DemoInterpFactory.__name__,
    },
}
tree_path_interp_factory_config_desc = TypeDesc(
    dict_schema = TreePathInterpFactoryConfigSchema(),
    ref_name = TreePathInterpFactoryConfigSchema.__name__,
    dict_example = tree_path_interp_factory_config_example,
    default_file_path = "",
)



from argrelay.custom_integ.DemoInterp import DemoInterp
from argrelay.custom_integ.DemoInterpFactoryConfigSchema import demo_interp_factory_config_desc
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.runtime_context.InterpContext import InterpContext


class DemoInterpFactory(AbstractInterpFactory):

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )
        demo_interp_factory_config_desc.validate_dict(config_dict)

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> DemoInterp:
        return DemoInterp(
            self.plugin_instance_id,
            self.config_dict,
            interp_ctx,
        )



import importlib
from typing import Type


class AbstractPlugin:

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        self.plugin_instance_id: str = plugin_instance_id
        self.config_dict: dict = config_dict

    def activate_plugin(self):
        """
        One-time plugin activation callback during server startup.

        This is a chance to initialize (once) any resources plugin requires.
        """
        pass

    def validate_loaded_data(self, static_data: "StaticData"):
        """
        Callback to validate data after all :class:`AbstractLoader`-s completed loading data.

        This is a chance for any plugin to verify that the data still matches its expectation.
        """
        pass


def import_plugin_class(plugin_entry):
    plugin_module = importlib.import_module(plugin_entry.plugin_module_name)
    plugin_class: Type[AbstractPlugin] = getattr(
        plugin_module,
        plugin_entry.plugin_class_name,
    )
    return plugin_class


def instantiate_plugin(
    plugin_instance_id: str,
    plugin_entry,
):
    plugin_class = import_plugin_class(plugin_entry)
    plugin_object: AbstractPlugin = plugin_class(
        plugin_instance_id,
        plugin_entry.plugin_config,
    )
    return plugin_object



from enum import Enum


class TermColor(Enum):
    """
    Color codes for terminal text
    """

    DARK_BLACK = "\033[30m"
    DARK_RED = "\033[31m"
    DARK_GREEN = "\033[32m"
    DARK_YELLOW = "\033[33m"
    DARK_BLUE = "\033[34m"
    DARK_MAGENTA = "\033[35m"
    DARK_CYAN = "\033[36m"
    DARK_WHITE = "\033[37m"

    BRIGHT_BLACK = "\033[90m"
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"

    DEBUG = BRIGHT_BLACK
    INFO = BRIGHT_GREEN

    WARNING = BRIGHT_YELLOW
    FAIL = BRIGHT_RED

    RESET = "\033[0m"



from __future__ import annotations

import copy
import os
import subprocess

from git import Repo

from argrelay.custom_integ.GitRepoArgType import GitRepoArgType
from argrelay.custom_integ.GitRepoDelegator import GitRepoDelegator, repo_root_abs_path_
from argrelay.custom_integ.GitRepoEntryConfigSchema import (
    repo_rel_path_,
    envelope_properties_,
    is_repo_enabled_,
)
from argrelay.custom_integ.GitRepoEnvelopeClass import GitRepoEnvelopeClass
from argrelay.custom_integ.GitRepoLoaderConfigSchema import (
    git_repo_loader_config_desc,
    is_plugin_enabled_,
    load_repo_commits_,
    repo_entries_,
)
from argrelay.custom_integ.value_constants import goto_repo_func_, desc_commit_func_
from argrelay.enum_desc.GlobalArgType import GlobalArgType
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper import eprint
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_id_,
    envelope_payload_,
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
)
from argrelay.schema_config_interp.SearchControlSchema import keys_to_types_list_, envelope_class_


class GitRepoLoader(AbstractLoader):
    """
    Implements FS_67_16_61_97 git_plugin.
    """

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )
        git_repo_loader_config_desc.validate_dict(config_dict)
        self.config_dict = git_repo_loader_config_desc.from_input_dict(config_dict)

    def update_static_data(self, static_data: StaticData) -> StaticData:
        """
        Scan `base_path` recursively and load metadata about all Git repos found.
        """

        if not self.config_dict:
            return static_data
        if not self.config_dict[is_plugin_enabled_]:
            return static_data

        static_data = self.load_git_objects(static_data)
        static_data = self.load_git_funcs(static_data)

        return static_data

    def load_git_objects(self, static_data: StaticData) -> StaticData:

        data_envelopes = static_data.data_envelopes

        # Init type keys (if they do not exist):
        for type_name in [enum_item.name for enum_item in GitRepoArgType]:
            if type_name not in static_data.known_arg_types:
                static_data.known_arg_types.append(type_name)

        # List of registered Git abs paths:
        repo_root_abs_paths = []
        for repo_base_abs_path in self.config_dict[repo_entries_]:

            repo_entries = self.config_dict[repo_entries_][repo_base_abs_path]
            repo_base_abs_path = os.path.expanduser(repo_base_abs_path)

            # Process repo entries collecting `repo_root_rel_path` and `repo_root_abs_path`:
            for repo_entry in repo_entries:

                repo_root_rel_path = repo_entry[repo_rel_path_]
                repo_root_abs_path = os.path.join(repo_base_abs_path, repo_root_rel_path)

                if not repo_entry[is_repo_enabled_]:
                    eprint(f"INFO: disabled repo: {repo_root_abs_path}")
                    continue

                # Deduplicate Git root paths:
                if repo_root_abs_path not in repo_root_abs_paths:
                    # Register `repo_root_abs_path`:
                    repo_root_abs_paths.append(repo_root_abs_path)
                else:
                    eprint(f"ERROR: duplicate Git repo: {repo_root_abs_path}")
                    raise RuntimeError

                # Query Git root path of curr dir:
                subproc = subprocess.run(
                    [
                        "git",
                        "-C",
                        repo_root_abs_path,
                        "rev-parse",
                        "--show-toplevel",
                    ],
                    capture_output = True,
                )
                ret_code = subproc.returncode

                if ret_code == 0:
                    pass
                else:
                    eprint(f"ERROR: not a Git repo: {repo_root_abs_path}")
                    # Walked into dir outside of Git repo:
                    continue

                repo_root_base_name = os.path.basename(repo_root_abs_path)

                # Produce path component list, for example, if:
                # repo_root_rel_path = "rel/path/to/git/repo.git/"
                # then:
                # path_comp_list = [ "rel", "path", "to", "git", "repo" ]
                path_comp_list = []
                for rel_git_path_comp in repo_root_rel_path.split(os.sep)[:-1]:
                    if rel_git_path_comp not in path_comp_list:
                        path_comp_list.append(rel_git_path_comp)

                ############################################################################################################
                # repos

                repo_envelope: dict = copy.deepcopy(repo_entry[envelope_properties_])

                repo_envelope.update({
                    envelope_id_: repo_root_abs_path,
                    envelope_payload_: {
                        repo_root_abs_path_: repo_root_abs_path,
                    },
                    ReservedArgType.EnvelopeClass.name: GitRepoEnvelopeClass.ClassGitRepo.name,
                    GlobalArgType.ObjectSelector.name: "repo",
                    GitRepoArgType.GitRepoRootRelPath.name: repo_root_rel_path,
                    GitRepoArgType.GitRepoRootAbsPath.name: repo_root_abs_path,
                    GitRepoArgType.GitRepoRootBaseName.name: repo_root_base_name,
                    GitRepoArgType.GitRepoPathComp.name: path_comp_list,
                })
                print(repo_envelope)
                data_envelopes.append(repo_envelope)

                if not self.config_dict[load_repo_commits_]:
                    continue

                git_repo = Repo(repo_root_abs_path)
                for git_commit in git_repo.iter_commits():
                    ########################################################################################################
                    # commits

                    commit_envelope: dict = copy.deepcopy(repo_entry[envelope_properties_])

                    commit_envelope.update({
                        envelope_id_: f"{repo_root_abs_path}:{git_commit.hexsha}",
                        envelope_payload_: {
                        },
                        ReservedArgType.EnvelopeClass.name: GitRepoEnvelopeClass.ClassGitCommit.name,
                        GlobalArgType.ObjectSelector.name: "commit",
                        GitRepoArgType.GitRepoRootRelPath.name: repo_root_rel_path,
                        GitRepoArgType.GitRepoRootAbsPath.name: repo_root_abs_path,
                        GitRepoArgType.GitRepoRootBaseName.name: repo_root_base_name,
                        GitRepoArgType.GitRepoPathComp.name: path_comp_list,
                        GitRepoArgType.GitRepoCommitId.name: git_commit.hexsha,
                        GitRepoArgType.GitRepoCommitAuthorName.name: git_commit.author.name,
                        GitRepoArgType.GitRepoCommitAuthorEmail.name: git_commit.author.email,
                        GitRepoArgType.GitRepoCommitMessage.name: git_commit.message,
                    })
                    data_envelopes.append(commit_envelope)

        return static_data

    # noinspection PyMethodMayBeStatic
    def load_git_funcs(self, static_data: StaticData) -> StaticData:

        data_envelopes = static_data.data_envelopes

        ###############################################################################################################
        # functions

        repo_search_control = {
            envelope_class_: GitRepoEnvelopeClass.ClassGitRepo.name,
            keys_to_types_list_: [
                {"alias": GitRepoArgType.GitRepoAlias.name},
                {"content": GitRepoArgType.GitRepoContentType.name},
                {"name": GitRepoArgType.GitRepoRootBaseName.name},
                {"path": GitRepoArgType.GitRepoRootRelPath.name},
                {"base": GitRepoArgType.GitRepoRootAbsPath.name},
                {"part": GitRepoArgType.GitRepoPathComp.name},
            ],
        }

        commit_search_control = {
            envelope_class_: GitRepoEnvelopeClass.ClassGitCommit.name,
            keys_to_types_list_: [
                {"name": GitRepoArgType.GitRepoRootBaseName.name},
                {"path": GitRepoArgType.GitRepoRootRelPath.name},
                {"base": GitRepoArgType.GitRepoRootAbsPath.name},
                {"email": GitRepoArgType.GitRepoCommitAuthorEmail.name},
                {"hex": GitRepoArgType.GitRepoCommitId.name},
            ],
        }

        given_function_envelope = {
            envelope_id_: goto_repo_func_,
            instance_data_: {
                delegator_plugin_instance_id_: GitRepoDelegator.__name__,
                search_control_list_: [
                    repo_search_control,
                ],
            },
            ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedArgType.HelpHint.name: "Describe Git repository",
            GlobalArgType.FunctionCategory.name: "external",
            GlobalArgType.ActionType.name: "goto",
            GlobalArgType.ObjectSelector.name: "repo",
        }
        data_envelopes.append(given_function_envelope)

        given_function_envelope = {
            envelope_id_: desc_commit_func_,
            instance_data_: {
                delegator_plugin_instance_id_: GitRepoDelegator.__name__,
                search_control_list_: [
                    commit_search_control,
                ],
            },
            ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedArgType.HelpHint.name: "Describe Git commit",
            GlobalArgType.FunctionCategory.name: "external",
            GlobalArgType.ActionType.name: "desc",
            GlobalArgType.ObjectSelector.name: "commit",
        }
        data_envelopes.append(given_function_envelope)

        return static_data



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.runtime_data.PluginEntry import PluginEntry


# TODO: Combine InterpResult and InvocationInput - only one need to exits
@dataclass
class InvocationInput:
    """
    See :class:`InvocationInputSchema`
    """

    all_tokens: list[str] = field()
    """
    Copy from `ParsedContext.all_tokens` - command line args.
    """

    consumed_tokens: list[int] = field()
    """
    Copy from `InterpContext.consumed_tokens` -
    indexes into `all_tokens` pointing to tokens consumed during interpretation.
    """

    delegator_plugin_entry: PluginEntry = field()
    """
    The `PluginEntry` taken from config on the server side to let client invoke that plugin.

    It is assumed that server and client code, if not of the same version, at least, compatible.
    """

    data_envelopes: list[dict] = field()
    """
    Envelopes copied from `InterpContext` at the end of command line interpretation on the server side.
    """

    custom_plugin_data: dict = field()
    """
    A placeholder dict for exclusive use by plugin.

    Whatever plugin server side needs to tell its plugin peer on the client side.
    """



from __future__ import annotations

from dataclasses import field, dataclass

from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper import eprint
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_server.HelpHintCache import HelpHintCache
from argrelay.relay_server.QueryEngine import QueryEngine, populate_query_dict
from argrelay.relay_server.QueryResult import QueryResult
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.runtime_context.ParsedContext import ParsedContext
from argrelay.runtime_context.SearchControl import SearchControl

function_envelope_ipos_ = 0


@dataclass
class InterpContext:
    """
    Mutable state for the process of command line interpretation.
    """

    parsed_ctx: ParsedContext = field()

    # TODO: Move all dynamic and non-serializable objects into `InterpRuntime` (or something like that).
    interp_factories: dict[str, "AbstractInterpFactory"] = field()
    """
    Reference to `ServerConfig.action_delegators`.
    """

    action_delegators: dict[str, "AbstractDelegator"] = field()
    """
    Reference to `ServerConfig.action_delegators`.
    """

    query_engine: QueryEngine = field()

    help_hint_cache: HelpHintCache = field()

    unconsumed_tokens: list[int] = field(init = False)
    """
    Remaining tokens (their ipos) which are still unconsumed (in ascending order).
    """

    consumed_tokens: list[int] = field(init = False, default_factory = lambda: [])
    """
    Already consumed tokens (their ipos) in the order of their consumption.
    """

    envelope_containers: list[EnvelopeContainer] = field(init = False, default_factory = lambda: [])
    """
    Each `envelope_container` wraps `data_envelope`-s matching query and some associated data.
    """

    curr_container: EnvelopeContainer = field(init = False, default = None)
    """
    One of the `envelope_containers` currently being searched.
    """

    curr_container_ipos: int = field(init = False, default = -1)
    """
    It is an `ipos` into `envelope_containers` to select `curr_container` (what is currently being searched for).

    Check `EnvelopeContainer.found_count` to see whether this container is singled out or not.

    All `envelope_containers` after `curr_container_ipos` will not be pointing to any `data_envelope`-s
    (search procedure has not reached them yet).
    """

    prev_interp: "AbstractInterp" = field(init = False, default = None)

    curr_interp: "AbstractInterp" = field(init = False, default = None)
    """
    Current interpreter during command line interpretation.
    """

    comp_suggestions: list[str] = field(init = False, default_factory = lambda: [])

    def __post_init__(self):
        self.unconsumed_tokens = self._init_unconsumed_tokens()

    def _init_unconsumed_tokens(self):
        if self.parsed_ctx.server_action == ServerAction.ProposeArgValues:
            return [
                token_ipos for token_ipos in range(0, len(self.parsed_ctx.all_tokens))
                # `ServerAction.ProposeArgValues` excludes tangent token because it is supposed to be completed:
                if token_ipos != self.parsed_ctx.tan_token_ipos
            ]
        else:
            return [
                token_ipos for token_ipos in range(0, len(self.parsed_ctx.all_tokens))
                # FS_23_62_89_43:
                # Process all tokens (including tangent token) in case of `ServerAction.DescribeLineArgs`.
                # Obviously, same applies for `ServerAction.RelayLineArgs` (as there is no chance to propose more).
            ]

    def alloc_searchable_containers(
        self,
        search_control_list: list[SearchControl],
    ):
        for search_control in search_control_list:
            envelope_container = EnvelopeContainer(search_control)
            self.envelope_containers.append(envelope_container)

    def is_last_container(self) -> bool:
        return self.curr_container_ipos + 1 == len(self.envelope_containers)

    def is_funct_found(self) -> bool:
        return (
            self.curr_container_ipos >= function_envelope_ipos_
            and
            self.envelope_containers[function_envelope_ipos_].found_count == 1
        )

    def query_prop_values(self):
        if not (
            self.curr_container
            and
            self.curr_container.search_control.envelope_class
        ):
            # If `ReservedArgType.EnvelopeClass` is not specified, nothing to search:
            return

        ElapsedTime.measure(f"begin_query_envelopes: {self.curr_container.search_control.envelope_class}")
        query_dict = populate_query_dict(self.curr_container)

        # TODO: FS_06_99_43_60: How to query values contained in arrays? For example, `GitRepoRootRelPath` is array. How to query envelopes which contain given value in elements of the array?
        query_result: QueryResult = self.query_engine.query_prop_values(
            query_dict,
            self.curr_container.search_control,
            self.curr_container.assigned_types_to_values,
        )
        self.curr_container.remaining_types_to_values = query_result.remaining_types_to_values
        self.curr_container.data_envelope = query_result.data_envelope
        self.curr_container.found_count = query_result.found_count
        ElapsedTime.measure(f"end_query_envelopes: {query_dict} {self.curr_container.found_count}")

    # TODO: Move all dynamic and non-serializable objects into `InterpRuntime` (or something like that) together with this loop:
    def interpret_command(self, first_interp_factory_id: str) -> None:
        """
        Main interpretation loop.

        Start with initial interpreter and continue until curr interpreter returns no more next interpreter.
        """

        interp_n: int = 0
        self.curr_interp = self.create_next_interp(first_interp_factory_id)
        while True:
            interp_n += 1

            # Query envelope values only - they will be used for consumption of command line args:
            ElapsedTime.measure(f"[i={interp_n}]: before_init_query: {type(self.curr_interp).__name__}")
            self.query_prop_values()

            ElapsedTime.measure(f"[i={interp_n}]: before_consume_args: {type(self.curr_interp).__name__}")
            self.curr_interp.consume_key_args()
            self.curr_interp.consume_pos_args()

            ElapsedTime.measure(f"[i={interp_n}]: before_reduce_query: {type(self.curr_interp).__name__}")
            self.query_prop_values()
            if self.curr_container:
                self.curr_container.populate_implicit_arg_values()

            if self.parsed_ctx.server_action == ServerAction.DescribeLineArgs:
                # Describing args will need to show options except default - query values before defaults:
                # TODO:
                pass

            # Apply defaults:
            self.curr_interp.run_fill_control()

            # Query envelopes after all defaults applied:
            # TODO: We could probably select whether to query only envelopes or their values depending on RunMode.
            #       But init of next envelope depends on prev envelope found.
            ElapsedTime.measure(f"[i={interp_n}]: before_final_query: {type(self.curr_interp).__name__}")
            self.query_prop_values()

            ElapsedTime.measure(f"[i={interp_n}]: before_try_iterate: {type(self.curr_interp).__name__}")
            interp_step: InterpStep = self.curr_interp.try_iterate()
            ElapsedTime.measure(f"[i={interp_n}]: after_try_iterate: {type(self.curr_interp).__name__}: {interp_step}")
            if interp_step == InterpStep.NextEnvelope:
                continue
            elif interp_step == InterpStep.StopAll:
                ElapsedTime.measure(
                    f"[i={interp_n}]: before_contribute_to_completion: {type(self.curr_interp).__name__}"
                )
                self._contribute_to_completion()
                ElapsedTime.measure(
                    f"[i={interp_n}]: after_contribute_to_completion: {type(self.curr_interp).__name__}"
                )
                return
            elif interp_step == InterpStep.NextInterp:
                ElapsedTime.measure(
                    f"[i={interp_n}]: before_contribute_to_completion: {type(self.curr_interp).__name__}"
                )
                self._contribute_to_completion()
                ElapsedTime.measure(
                    f"[i={interp_n}]: after_contribute_to_completion: {type(self.curr_interp).__name__}"
                )
                self.prev_interp = self.curr_interp
                next_interp = self.curr_interp.next_interp()
                if next_interp:
                    self.curr_interp = next_interp
                else:
                    return
            else:
                raise RuntimeError(interp_step)

    def _contribute_to_completion(self):
        if self.parsed_ctx.server_action == ServerAction.ProposeArgValues:
            # Each in the chains of interpreters hava a chance to suggest completion values (contribute):
            self.curr_interp.propose_arg_completion()

    def propose_arg_values(self) -> list[str]:
        """
        FS_71_87_33_52: remove `help_hint` (after first space) if there is only one option
        """
        if len(self.comp_suggestions) == 1:
            first_space_ipos = self.comp_suggestions[0].find(' ')
            if first_space_ipos >= 0:
                self.comp_suggestions[0] = self.comp_suggestions[0][:first_space_ipos]
        return self.comp_suggestions

    def create_next_interp(
        self,
        interp_factory_id: str,
    ) -> "AbstractInterp":
        interp_factory: "AbstractInterpFactory" = self.interp_factories[interp_factory_id]
        return interp_factory.create_interp(
            self,
        )

    def print_debug(self, end_str: str = "\n") -> None:
        if not self.parsed_ctx.is_debug_enabled:
            return
        self.parsed_ctx.print_debug("")
        eprint(TermColor.DEBUG.value, end = "")
        eprint(f"comp_suggestions: {self.comp_suggestions}", end = " ")
        eprint(TermColor.RESET.value, end = end_str)



from __future__ import annotations

from argrelay.custom_integ.ServiceArgType import ServiceArgType
from argrelay.custom_integ.value_constants import (
    goto_host_funct_,
    goto_service_funct_,
    list_service_func_,
    list_host_func_,
)
from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.plugin_delegator.AbstractDelegator import (
    AbstractDelegator,
    get_data_envelopes,
    get_func_name_from_container,
    get_func_name_from_envelope,
)
from argrelay.plugin_delegator.ErrorDelegator import ErrorDelegator
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import (
    error_message_,
    error_code_,
    error_delegator_custom_data_desc,
    error_delegator_stub_custom_data_example,
)
from argrelay.plugin_delegator.InvocationInput import InvocationInput
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.relay_server.QueryEngine import populate_query_dict
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.AssignedValue import AssignedValue

host_envelope_ipos_ = 1
service_envelope_ipos_ = 1
access_envelope_ipos_ = 2


def set_default_to(arg_type, arg_val, envelope_container):
    if arg_type in envelope_container.search_control.types_to_keys_dict.keys():
        if arg_type not in envelope_container.assigned_types_to_values.keys():
            if arg_type in envelope_container.remaining_types_to_values:
                if arg_val in envelope_container.remaining_types_to_values[arg_type]:
                    del envelope_container.remaining_types_to_values[arg_type]
                    envelope_container.assigned_types_to_values[arg_type] = AssignedValue(
                        arg_val,
                        ArgSource.DefaultValue,
                    )
        else:
            # Setting `ArgSource.DefaultValue`: it cannot override any, right? No point to handle assigned case:
            pass


def redirect_to_no_func_error(
    interp_ctx,
    server_config,
):
    return redirect_to_error(
        interp_ctx,
        server_config,
        "ERROR: objects cannot be searched until function is fully qualified",
        1,
    )


def redirect_to_error(
    interp_ctx,
    server_config,
    error_message,
    error_code,
):
    # Redirect to `ErrorDelegator`:
    delegator_plugin_instance_id = ErrorDelegator.__name__
    custom_plugin_data = {
        error_message_: error_message,
        error_code_: error_code,
    }
    error_delegator_custom_data_desc.validate_dict(custom_plugin_data)
    invocation_input = InvocationInput(
        all_tokens = interp_ctx.parsed_ctx.all_tokens,
        consumed_tokens = interp_ctx.consumed_tokens,
        delegator_plugin_entry = server_config.plugin_dict[delegator_plugin_instance_id],
        data_envelopes = get_data_envelopes(interp_ctx),
        custom_plugin_data = custom_plugin_data,
    )
    return invocation_input


class ServiceDelegator(AbstractDelegator):

    def __init__(
        self,
        plugin_instance_id: str,
        config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            config_dict,
        )

    def run_fill_control(
        self,
        interp_ctx: "InterpContext",
    ):
        func_name = get_func_name_from_container(interp_ctx)
        if func_name in [
            goto_host_funct_,
            goto_service_funct_,
        ]:
            assert host_envelope_ipos_ == service_envelope_ipos_
            object_envelope_ipos = host_envelope_ipos_

            # If need to specify `AccessType` `data_envelope`:
            if interp_ctx.curr_container_ipos == interp_ctx.curr_interp.base_envelope_ipos + access_envelope_ipos_:
                # Take object found so far:
                object_envelope = interp_ctx.envelope_containers[(
                    interp_ctx.curr_interp.base_envelope_ipos + object_envelope_ipos
                )].data_envelope

                access_container = interp_ctx.envelope_containers[(
                    interp_ctx.curr_interp.base_envelope_ipos + access_envelope_ipos_
                )]

                # Select default value to search `AccessType` `data_envelope` based on `CodeMaturity`:
                code_arg_type = ServiceArgType.CodeMaturity.name
                if code_arg_type in object_envelope:
                    code_arg_val = object_envelope[code_arg_type]
                    if code_arg_val == "prod":
                        set_default_to(ServiceArgType.AccessType.name, "ro", access_container)
                    else:
                        set_default_to(ServiceArgType.AccessType.name, "rw", access_container)

        elif func_name in [
            list_host_func_,
            list_service_func_,
        ]:
            pass
        else:
            raise RuntimeError

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_funct_found(), "the (first) function envelope must be found"

        func_name = get_func_name_from_container(interp_ctx)

        if func_name in [
            goto_host_funct_,
            goto_service_funct_,
        ]:
            # Actual implementation is not defined for demo:
            return redirect_to_error(
                interp_ctx,
                local_server.server_config,
                error_delegator_stub_custom_data_example[error_message_],
                error_delegator_stub_custom_data_example[error_code_],
            )
        elif func_name in [
            list_host_func_,
            list_service_func_,
        ]:
            vararg_data_envelope_ipos = host_envelope_ipos_
            assert vararg_data_envelope_ipos == host_envelope_ipos_ == service_envelope_ipos_
            # Verify that func is selected and all what is left to do is to query 0...N objects:
            if interp_ctx.curr_container_ipos >= vararg_data_envelope_ipos:
                # Search `data_envelope`-s based on existing args on command line:
                query_dict = populate_query_dict(interp_ctx.envelope_containers[vararg_data_envelope_ipos])
                # Plugin to invoke on client side:
                delegator_plugin_instance_id = ServiceDelegator.__name__
                # Package into `InvocationInput` payload object:
                invocation_input = InvocationInput(
                    all_tokens = interp_ctx.parsed_ctx.all_tokens,
                    consumed_tokens = interp_ctx.consumed_tokens,
                    delegator_plugin_entry = local_server.server_config.plugin_dict[delegator_plugin_instance_id],
                    data_envelopes = (
                        # existing envelopes (until vararg one):
                        get_data_envelopes(interp_ctx)[:vararg_data_envelope_ipos]
                        +
                        # all envelopes in vararg set:
                        local_server.get_query_engine().query_data_envelopes(query_dict)
                    ),
                    custom_plugin_data = {},
                )
                return invocation_input
            else:
                return redirect_to_no_func_error(
                    interp_ctx,
                    local_server.server_config,
                )
        else:
            # Plugin is given a function name it does not know:
            raise RuntimeError

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        """
        Print `data_envelope`-s received from server on client side.
        """

        func_name = get_func_name_from_envelope(invocation_input.data_envelopes)
        if func_name == list_host_func_:
            for data_envelope in invocation_input.data_envelopes[host_envelope_ipos_:]:
                print(data_envelope)
        elif func_name == list_service_func_:
            for data_envelope in invocation_input.data_envelopes[service_envelope_ipos_:]:
                print(data_envelope)
        else:
            raise RuntimeError



import logging

import pkg_resources
from flasgger import Swagger
from flask import Flask, request, redirect

from argrelay import relay_server
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.relay_server.route_api import create_blueprint_api
from argrelay.relay_server.route_gui import create_blueprint_gui
from argrelay.schema_config_core_server.ServerConfigSchema import server_config_desc
from argrelay.server_spec.const_int import API_SPEC_PATH, API_DOCS_PATH, ARGRELAY_GUI_PATH

# Set this here (because `require` function may fail in other contexts):
server_version = pkg_resources.require("argrelay")[0].version
server_title = relay_server.__name__


class CustomFlaskApp(Flask):
    """
    This is an API-wrapper exposing `LocalServer` over the network.
    """

    def __init__(
        self,
        import_name: str,
        static_url_path = "/gui_static",
        static_folder = pkg_resources.resource_filename(relay_server.__name__, "gui_static"),
        template_folder = pkg_resources.resource_filename(relay_server.__name__, "gui_templates"),
    ):
        super().__init__(
            import_name = import_name,
            static_folder = static_folder,
            template_folder = template_folder,
            static_url_path = static_url_path,
        )
        self.local_server: LocalServer = LocalServer(server_config_desc.from_default_file())

    def run_with_config(self):
        # Use custom logging at DEBUG level - see: `log_request` and `log_response:
        self.logger.setLevel(logging.DEBUG)
        # Log only at ERROR level by default logger:
        # https://stackoverflow.com/a/18379764/441652
        logging.getLogger("werkzeug").setLevel(logging.ERROR)

        self.run(
            # Contrary to "debug" keyword, if IDE debug is needed, set `debug` to `False` to avoid reloader:
            # https://stackoverflow.com/a/53790400/441652
            # NOTE: Disabled debug to make sure reloader does not restart server (which does not work with Mongo DB):
            debug = False,
            host = self.local_server.server_config.connection_config.server_host_name,
            port = self.local_server.server_config.connection_config.server_port_number,
        )


def create_app() -> CustomFlaskApp:
    """
    The function prevents running the code inside it on import - necessary for pre-test mocking.
    Note that Flask may rely on specific name `create_app` to find this factory:
    https://flask.palletsprojects.com/en/2.2.x/patterns/appfactories/#using-applications
    """

    flask_app = CustomFlaskApp(
        import_name = relay_server.__name__,
    )

    flask_app.local_server.start_local_server()

    swagger_template = {
        "info": {
            "title": server_title,
            "version": server_version,
            "description": f"See <a href=\"{ARGRELAY_GUI_PATH}\" target=\"_blank\">built-in GUI</a>.",
        },
    }

    swagger_config = {
        "headers": [],
        "specs": [
            {
                "endpoint": "relay_server",
                "route": API_SPEC_PATH,
                "rule_filter": lambda rule: True,
                "model_filter": lambda tag: True,
            },
        ],
        "static_url_path": "/flasgger_static",
        "swagger_ui": True,
        "specs_route": API_DOCS_PATH,
    }

    Swagger(
        flask_app,
        template = swagger_template,
        config = swagger_config,
    )

    @flask_app.before_request
    def log_request():
        flask_app.logger.debug(
            "request: %s\n    source: %s\n    headers: %s\n    body: %s",
            request,
            request.remote_addr,
            request.headers.__repr__(),
            request.get_data(),
        )

    @flask_app.after_request
    def log_response(response):
        # Wrap into try/except because of this:
        # https://stackoverflow.com/q/64006669/441652
        # noinspection PyBroadException
        try:
            response_data = str(response.get_data())
        except Exception as e:
            response_data = str(e)
        flask_app.logger.debug(
            "response: %s\n    headers: %s\n    body: %s\n",
            response,
            response.headers.__repr__(),
            response_data,
        )
        return response

    @flask_app.route("/")
    def root_redirect():
        return redirect(ARGRELAY_GUI_PATH, code = 302)

    flask_app.register_blueprint(create_blueprint_api(flask_app.local_server))
    flask_app.register_blueprint(create_blueprint_gui(
        server_version,
        flask_app.local_server.server_config.gui_banner_config,
        flask_app.local_server.server_start_time,
    ))

    return flask_app



from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.schema_config_interp.SearchControlSchema import search_control_desc

delegator_plugin_instance_id_ = "delegator_plugin_instance_id"
search_control_list_ = "search_control_list"


class FunctionEnvelopeInstanceDataSchema(Schema):
    """
    Schema for :class:`DataEnvelopeSchema.instance_data` of :class:`ReservedEnvelopeClass.ClassFunction`
    """

    class Meta:
        unknown = RAISE
        strict = True

    delegator_plugin_instance_id = fields.String(
        required = True,
    )

    search_control_list = fields.List(
        fields.Nested(search_control_desc.dict_schema),
        required = True,
    )


function_envelope_instance_data_desc = TypeDesc(
    dict_schema = FunctionEnvelopeInstanceDataSchema(),
    ref_name = FunctionEnvelopeInstanceDataSchema.__name__,
    dict_example = {
        delegator_plugin_instance_id_: "NoopDelegator",
        search_control_list_: [
            search_control_desc.dict_example,
        ],
    },
    default_file_path = "",
)



from marshmallow import Schema, fields, RAISE, post_load

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.mongo_data.MongoServerConfig import MongoServerConfig

database_name_ = "database_name"
start_server_ = "start_server"
server_start_command_ = "server_start_command"


class MongoServerConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    database_name = fields.String()

    start_server = fields.Boolean()

    server_start_command = fields.String()

    @post_load
    def make_object(self, input_dict, **kwargs):
        return MongoServerConfig(
            database_name = input_dict[database_name_],
            start_server = input_dict[start_server_],
            server_start_command = input_dict[server_start_command_],
        )


mongo_server_config_desc = TypeDesc(
    dict_schema = MongoServerConfigSchema(),
    ref_name = MongoServerConfigSchema.__name__,
    dict_example = {
        database_name_: "argrelay",
        start_server_: False,
        server_start_command_:
            "~/argrelay.git/temp/mongo/mongodb-linux-x86_64-rhel80-6.0.3/bin/mongod"
            " --dbpath "
            "~/argrelay.git/temp/mongo/data",
    },
    default_file_path = "",
)



from __future__ import annotations

from copy import deepcopy

import mongomock
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database

from argrelay.misc_helper import eprint
from argrelay.mongo_data.MongoConfig import MongoConfig
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_server.StaticDataSchema import data_envelopes_
from argrelay.schema_config_interp.DataEnvelopeSchema import envelope_id_, data_envelope_desc, mongo_id_


def get_mongo_client(mongo_config: MongoConfig):
    if mongo_config.use_mongomock_only:
        return mongomock.MongoClient()
    else:
        return MongoClient(mongo_config.mongo_client.client_connection_string)


def store_envelopes(mongo_db: Database, static_data: StaticData):
    # TODO: Currently, we use single collection,
    #       but we never search across all object,
    #       we always search specific `ReservedArgType.EnvelopeClass.name`.
    #       Should we use collection per envelope class?
    #       However, what if we need to list all envelop classes?
    #       Then, single collection is fine (and simple).
    col_proxy: Collection = mongo_db[data_envelopes_]
    col_proxy.delete_many({})

    envelope_i: int = 0
    envelope_n = len(static_data.data_envelopes)
    for data_envelope in static_data.data_envelopes:
        if envelope_i % 1_000 == 0:
            eprint(f"indexed envelopes {envelope_i}/{envelope_n}...")
        envelope_i += 1
        envelope_to_store = deepcopy(data_envelope)
        data_envelope_desc.validate_dict(envelope_to_store)
        if envelope_id_ in envelope_to_store:
            envelope_to_store[mongo_id_] = data_envelope[envelope_id_]
        col_proxy.insert_one(envelope_to_store)


def create_index(mongo_db: Database, static_data: StaticData):
    col_proxy: Collection = mongo_db[data_envelopes_]

    for index_field in static_data.known_arg_types:
        col_proxy.create_index(index_field)


def find_envelopes(mongo_db: Database, assigned_types_to_values: dict[str, AssignedValue]):
    col_proxy: Collection = mongo_db[data_envelopes_]
    query_dict = {}
    for arg_type, arg_val in assigned_types_to_values.items():
        query_dict[arg_type] = arg_val.arg_value

    return col_proxy.find(query_dict)



from marshmallow import Schema, fields, RAISE, post_load

from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.relay_server.GuiBannerConfig import GuiBannerConfig

header_html_ = "header_html"
footer_html_ = "footer_html"


class GuiBannerConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    header_html = fields.String()
    footer_html = fields.String()

    @post_load
    def make_object(self, input_dict, **kwargs):
        return GuiBannerConfig(
            header_html = input_dict[header_html_],
            footer_html = input_dict[footer_html_],
        )


gui_banner_config_desc = TypeDesc(
    dict_schema = GuiBannerConfigSchema(),
    ref_name = GuiBannerConfigSchema.__name__,
    dict_example = {
        header_html_: "",
        footer_html_: "",
    },
    default_file_path = "",
)



from argrelay.misc_helper.AbstractPlugin import AbstractPlugin
from argrelay.runtime_data.StaticData import StaticData


class AbstractLoader(AbstractPlugin):

    def update_static_data(
        self,
        static_data: StaticData,
    ) -> StaticData:
        pass



from __future__ import annotations

from argrelay.plugin_interp.FuncArgsInterp import FuncArgsInterp


class DemoInterp(FuncArgsInterp):
    """
    `DemoInterp` is a configured use case of :class:`FuncArgsInterp` to demo auto-completion.

    See FS_55_57_45_04.
    """

    pass



from argrelay.enum_desc.GlobalArgType import GlobalArgType
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.plugin_delegator.InterceptDelegator import InterceptDelegator
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_interp.DataEnvelopeSchema import envelope_id_, instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
)


class InterceptLoader(AbstractLoader):

    def update_static_data(
        self,
        static_data: StaticData,
    ) -> StaticData:
        data_envelopes = static_data.data_envelopes

        given_function_envelope = {
            envelope_id_: SpecialFunc.intercept_func.name,
            instance_data_: {
                delegator_plugin_instance_id_: InterceptDelegator.__name__,
                search_control_list_: [
                ],
            },
            ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedArgType.HelpHint.name: (
                "Intercept and print `InvocationPayload` "
                "for specified function and its args"
            ),
            GlobalArgType.FunctionCategory.name: "internal",
            GlobalArgType.ActionType.name: "intercept",
            GlobalArgType.ObjectSelector.name: "func",
        }
        data_envelopes.append(given_function_envelope)

        return static_data



from __future__ import annotations

import copy
import json

from cachetools import TTLCache
from pymongo.collection import Collection
from pymongo.database import Database

from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.misc_helper.ElapsedTime import ElapsedTime
from argrelay.relay_server.QueryCacheConfig import QueryCacheConfig
from argrelay.relay_server.QueryResult import QueryResult
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.schema_config_core_server.StaticDataSchema import data_envelopes_


class QueryEngine:

    def __init__(
        self,
        query_cache_config: QueryCacheConfig,
        mongo_db: Database,
    ):
        self.mongo_db: Database = mongo_db
        self.mongo_col: Collection = self.mongo_db[data_envelopes_]
        self.query_cache: TTLCache = TTLCache(
            maxsize = query_cache_config.query_cache_max_size_bytes,
            ttl = query_cache_config.query_cache_ttl_sec,
        )
        self.enable_query_cache: bool = query_cache_config.enable_query_cache

    def query_data_envelopes(
        self,
        query_dict: dict,
    ) -> list[dict]:
        """
        This query is used for final invocation for vararg-like multiple `data_envelope`-s (FS_18_64_57_18).
        Therefore, it is not latency-sensitive (results are not cached).
        """

        query_res = self.mongo_col.find(query_dict)
        return list(iter(query_res))

    def query_prop_values(
        self,
        query_dict: dict,
        search_control: SearchControl,
        assigned_types_to_values: dict[str, AssignedValue],
    ) -> QueryResult:
        """
        Query values per types in `search_control` and populates `remaining_types_to_values`.

        This query is used in completion which makes it latency-sensitive (results are cached).

        It also combines:
        *   counting total `found_count` of `data_envelope`-s returned and
        *   storing the last `data_envelope`.
        The last `data_envelope` is only useful when `found_count` is one (making it unambiguous `data_envelope`).
        To search all `data_envelope`, use `query_data_envelopes` function.
        """

        if self.enable_query_cache:
            ElapsedTime.measure("before_cache_lookup")
            query_key = json.dumps(query_dict, separators = (",", ":"))
            query_result = self.query_cache.get(query_key)
            ElapsedTime.measure("after_cache_lookup")
            if query_result:
                return copy.deepcopy(query_result)

            query_result = self._query_prop_values(
                assigned_types_to_values,
                query_dict,
                search_control,
            )

            self.query_cache[query_key] = copy.deepcopy(query_result)
        else:
            query_result = self._query_prop_values(
                assigned_types_to_values,
                query_dict,
                search_control,
            )
        # No cache -> no deep copy (throw away result):
        return query_result

    def _query_prop_values(
        self,
        assigned_types_to_values,
        query_dict,
        search_control,
    ) -> QueryResult:

        ElapsedTime.measure("before_mongo_find")
        query_res = self.mongo_col.find(query_dict)
        ElapsedTime.measure("after_mongo_find")
        query_result = self._process_prop_values(
            query_res,
            search_control,
            assigned_types_to_values,
        )
        ElapsedTime.measure("after_process_results")
        return query_result

    @staticmethod
    def _process_prop_values(
        query_res,
        search_control: SearchControl,
        assigned_types_to_values: dict[str, AssignedValue],
    ) -> QueryResult:
        """
        Populates:
        *   `found_count`
        *   `remaining_types_to_values`
        """

        remaining_types_to_values: dict[str, list[str]] = {}
        data_envelope = None
        found_count = 0

        # TODO: What if search result is huge? Blame data set designer?
        # find all remaining arg vals per arg type:
        for data_envelope in iter(query_res):
            found_count += 1
            # `arg_type` must be known:
            for arg_type in search_control.types_to_keys_dict.keys():
                # `arg_type` must be in one of the `data_envelope`-s found:
                if arg_type in data_envelope:
                    # If assigned/consumed, `arg_type` must not appear
                    # as an option in `remaining_types_to_values` again:
                    if arg_type not in assigned_types_to_values.keys():
                        arg_val = data_envelope[arg_type]
                        if arg_type not in remaining_types_to_values:
                            val_list = []
                            remaining_types_to_values[arg_type] = val_list
                        else:
                            val_list = remaining_types_to_values[arg_type]
                        # Deduplicate: ensure unique `arg_value`-s:
                        if arg_val not in val_list:
                            val_list.append(arg_val)

        return QueryResult(
            data_envelope,
            found_count,
            remaining_types_to_values,
        )


def populate_query_dict(envelope_container):
    query_dict = {
        ReservedArgType.EnvelopeClass.name: envelope_container.search_control.envelope_class,
    }
    # FS_31_70_49_15: populate arg values to search from the context:
    for arg_type in envelope_container.search_control.types_to_keys_dict.keys():
        if arg_type in envelope_container.assigned_types_to_values:
            query_dict[arg_type] = envelope_container.assigned_types_to_values[arg_type].arg_value
    return query_dict



from marshmallow import Schema, RAISE, fields, validates_schema, ValidationError, post_load

from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper.TypeDesc import TypeDesc
from argrelay.runtime_context.SearchControl import SearchControl

envelope_class_ = "envelope_class"
"""
Specifies `ReservedArgType.EnvelopeClass` to search.
"""

keys_to_types_list_ = "keys_to_types_list"
"""
List of keys to use for named args during interpretation and arg types to use in query.
"""


class SearchControlSchema(Schema):
    """
    Implements FS_31_70_49_15 # search_control

    Schema for search_control which specifies what arg types will be used in search of the `data_envelope`
    (and how they are mapped to named arg keys).
    """

    class Meta:
        unknown = RAISE
        strict = True

    envelope_class = fields.String()

    keys_to_types_list = fields.List(
        fields.Dict(
            keys = fields.String(),
            values = fields.String(),
            required = True,
        ),
        required = True,
    )

    @post_load
    def make_object(self, input_dict, **kwargs):
        return SearchControl(
            envelope_class = input_dict[envelope_class_],
            keys_to_types_list = input_dict[keys_to_types_list_],
        )

    @validates_schema
    def validate_known(self, input_dict, **kwargs):
        for key_to_type_entry in input_dict[keys_to_types_list_]:
            # ensure there is only one key per dict:
            if len(key_to_type_entry.keys()) != 1:
                raise ValidationError("only one key per dict is allowed")


search_control_desc = TypeDesc(
    dict_schema = SearchControlSchema(),
    ref_name = SearchControlSchema.__name__,
    dict_example = {
        envelope_class_: ReservedEnvelopeClass.ClassFunction.name,
        keys_to_types_list_: [
            {
                "type_a": "TypeA",
            },
            {
                "type_b": "TypeB",
            },
        ],
    },
    default_file_path = "",
)



from __future__ import annotations

from enum import Enum, auto


class GitRepoArgType(Enum):
    """
    Envelope arg types (properties) used by `GitRepoLoader`.
    """

    # Short name for a repo (to avoid specifying its `GitRepoRootRelPath`):
    GitRepoAlias = auto()

    # Absolute path common to multiple Git repos - what `GitRepoRootRelPath` are relative to:
    GitRepoRootAbsPath = auto()

    # Git root path relative to `GitRepoRootAbsPath`:
    GitRepoRootRelPath = auto()

    # Git root dir base name:
    GitRepoRootBaseName = auto()

    # Categorize repo content: conf, code, ...
    GitRepoContentType = auto()

    # TODO: FS_06_99_43_60: this is experimental:
    # Path component leading to Git repo root path (e.g. ["qwer", "asdf", "zxcv"] of "zxcv/qwer/asdf"):
    GitRepoPathComp = auto()

    # TODO: FS_06_99_43_60: this is experimental:
    # Git remote, for example: "git@github.com:uvsmtid/argrelay.git"
    GitRepoRemoteUrl = auto()

    # TODO: FS_06_99_43_60: this is experimental:
    # Git local branch name:
    GitRepoLocalBranch = auto()

    # Git commit, for example: 52a3d54d17fe5cba6bad57abc4b6bebd881d3ce6
    GitRepoCommitId = auto()

    # Git commit comment:
    GitRepoCommitMessage = auto()

    # Git commit author name:
    GitRepoCommitAuthorName = auto()

    # Git commit author email:
    GitRepoCommitAuthorEmail = auto()

    def __str__(self):
        return self.name
